"""
Intercom source implementation for data ingestion.

This module provides DLT sources for retrieving data from Intercom API endpoints
including contacts, companies, conversations, tickets, and more.
"""

from typing import Optional, Sequence

import dlt
from dlt.common.time import ensure_pendulum_datetime
from dlt.common.typing import TAnyDateTime
from dlt.sources import DltResource, DltSource

from .helpers import (
    IntercomAPIClient,
    IntercomCredentialsAccessToken,
    TIntercomCredentials,
    convert_datetime_to_timestamp,
    create_resource_from_config,
    transform_company,
    transform_contact,
    transform_conversation,
)
from .helpers import (
    IntercomCredentialsOAuth as IntercomCredentialsOAuth,
)
from .settings import (
    DEFAULT_START_DATE,
    RESOURCE_CONFIGS,
)


@dlt.source(name="intercom", max_table_nesting=0)
def intercom_source(
    credentials: TIntercomCredentials = dlt.secrets.value,
    start_date: Optional[TAnyDateTime] = DEFAULT_START_DATE,
    end_date: Optional[TAnyDateTime] = None,
) -> Sequence[DltResource]:
    """
    A DLT source that retrieves data from Intercom API.

    This source provides access to various Intercom resources including contacts,
    companies, conversations, tickets, and more. It supports incremental loading
    for resources that track updated timestamps.

    Args:
        credentials: Intercom API credentials (AccessToken or OAuth).
            Defaults to dlt.secrets.value.
        start_date: The start date for incremental loading.
            Defaults to January 1, 2020.
        end_date: Optional end date for incremental loading.
            If not provided, loads all data from start_date to present.

    Returns:
        Sequence of DLT resources for different Intercom endpoints.

    Example:
        >>> source = intercom_source(
        ...     credentials=IntercomCredentialsAccessToken(
        ...         access_token="your_token",
        ...         region="us"
        ...     ),
        ...     start_date=datetime(2024, 1, 1)
        ... )
    """
    # Initialize API client
    api_client = IntercomAPIClient(credentials)

    # Convert dates to pendulum and then to unix timestamps for Intercom API
    start_date_obj = ensure_pendulum_datetime(start_date) if start_date else None
    end_date_obj = ensure_pendulum_datetime(end_date) if end_date else None

    # Convert to unix timestamps for API compatibility
    # Use default start date if none provided
    if not start_date_obj:
        from .settings import DEFAULT_START_DATE

        start_date_obj = ensure_pendulum_datetime(DEFAULT_START_DATE)

    start_timestamp = convert_datetime_to_timestamp(start_date_obj)
    end_timestamp = (
        convert_datetime_to_timestamp(end_date_obj) if end_date_obj else None
    )

    # Transform function mapping
    transform_functions = {
        "transform_contact": transform_contact,
        "transform_company": transform_company,
        "transform_conversation": transform_conversation,
    }

    # Generate all resources from configuration
    resources = []
    for resource_name, config in RESOURCE_CONFIGS.items():
        resource_func = create_resource_from_config(
            resource_name,
            config,
            api_client,
            start_timestamp,
            end_timestamp,
            transform_functions,
        )

        # Call the resource function to get the actual resource
        resources.append(resource_func())

    return resources


def intercom(
    api_key: str,
    region: str = "us",
    start_date: Optional[TAnyDateTime] = DEFAULT_START_DATE,
    end_date: Optional[TAnyDateTime] = None,
) -> DltSource:
    """
    Convenience function to create Intercom source with access token.

    Args:
        api_key: Intercom API access token.
        region: Data region (us, eu, or au). Defaults to "us".
        start_date: Start date for incremental loading.
        end_date: Optional end date for incremental loading.

    Returns:
        Sequence of DLT resources.

    Example:
        >>> source = intercom(
        ...     api_key="your_access_token",
        ...     region="us",
        ...     start_date=datetime(2024, 1, 1)
        ... )
    """
    credentials = IntercomCredentialsAccessToken(access_token=api_key, region=region)

    return intercom_source(
        credentials=credentials,
        start_date=start_date,
        end_date=end_date,
    )



"""Trustpilot source for ingesting reviews."""

from typing import Any, Dict, Generator, Iterable

import dlt
import pendulum
from dlt.sources import DltResource

from .client import TrustpilotClient


@dlt.source()
def trustpilot_source(
    business_unit_id: str,
    start_date: str,
    end_date: str | None,
    api_key: str,
    per_page: int = 1000,
) -> Iterable[DltResource]:
    """Return resources for Trustpilot."""

    client = TrustpilotClient(api_key=api_key)

    @dlt.resource(name="reviews", write_disposition="merge", primary_key="id")
    def reviews(
        dateTime=(
            dlt.sources.incremental(
                "updated_at",
                initial_value=start_date,
                end_value=end_date,
                range_start="closed",
                range_end="closed",
            )
        ),
    ) -> Generator[Dict[str, Any], None, None]:
        if end_date is None:
            end_dt = pendulum.now(tz="UTC").isoformat()
        else:
            end_dt = dateTime.end_value
        start_dt = dateTime.last_value
        yield from client.paginated_reviews(
            business_unit_id=business_unit_id,
            per_page=per_page,
            updated_since=start_dt,
            end_date=end_dt,
        )

    yield reviews



"""Simple ClickUp source."""

from datetime import datetime
from typing import Iterable

import dlt
import pendulum
from dlt.common.time import ensure_pendulum_datetime
from dlt.sources import DltResource

from .helpers import ClickupClient


@dlt.source(max_table_nesting=0)
def clickup_source(
    api_token: str = dlt.secrets.value,
    start_date: datetime = None,
    end_date: datetime = None,
) -> Iterable[DltResource]:
    client = ClickupClient(api_token)

    @dlt.resource(
        name="user",
        primary_key="id",
        write_disposition="merge",
    )
    def user() -> Iterable[dict]:
        data = client.get("/user")
        yield data["user"]

    @dlt.resource(name="teams", primary_key="id", write_disposition="merge")
    def teams() -> Iterable[dict]:
        for team in client.get_teams():
            yield team

    @dlt.resource(name="spaces", primary_key="id", write_disposition="merge")
    def spaces() -> Iterable[dict]:
        for space in client.get_spaces():
            yield space

    @dlt.resource(name="lists", write_disposition="merge", primary_key="id")
    def lists() -> Iterable[dict]:
        for list in client.get_lists():
            yield list

    @dlt.resource(
        name="tasks",
        write_disposition="merge",
        primary_key="id",
        columns={"date_updated": {"data_type": "timestamp"}},
    )
    def tasks(
        date_updated: dlt.sources.incremental[str] = dlt.sources.incremental(
            "date_updated",
            initial_value=ensure_pendulum_datetime(start_date).in_timezone("UTC"),
            range_end="closed",
            range_start="closed",
        ),
    ) -> Iterable[dict]:
        if date_updated.last_value:
            start = ensure_pendulum_datetime(date_updated.last_value).in_timezone("UTC")
        else:
            start = ensure_pendulum_datetime(start_date).in_timezone("UTC")

        if date_updated.end_value is None:
            end = pendulum.now("UTC")
        else:
            end = date_updated.end_value.in_timezone("UTC")

        for list_obj in client.get_lists():
            for task in client.paginated(
                f"/list/{list_obj['id']}/task", "tasks", {"page_size": 100}
            ):
                task_dt = ensure_pendulum_datetime(int(task["date_updated"]) / 1000)
                if task_dt >= start and task_dt <= end:
                    task["date_updated"] = task_dt
                    yield task

    return (
        user,
        teams,
        spaces,
        lists,
        tasks,
    )



# Copyright 2022-2025 ScaleVector
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

RATE_LIMIT = """
  rateLimit {
    limit
    cost
    remaining
    resetAt
  }
"""

ISSUES_QUERY = """
query($owner: String!, $name: String!, $issues_per_page: Int!, $first_reactions: Int!, $first_comments: Int!, $page_after: String) {
  repository(owner: $owner, name: $name) {
    %s(first: $issues_per_page, orderBy: {field: CREATED_AT, direction: DESC}, after: $page_after) {
      totalCount
      pageInfo {
        endCursor
        startCursor
      }
      nodes {
        # id
        number
        url
        title
        body
        author {login avatarUrl url}
        authorAssociation
        closed
        closedAt
        createdAt
        state
        updatedAt
        reactions(first: $first_reactions) {
          totalCount
          nodes {
            # id
            user {login avatarUrl url}
            content
            createdAt
          }
        }
        comments(first: $first_comments) {
          totalCount
          nodes {
            id
            url
            body
            author {avatarUrl login url}
            authorAssociation
            createdAt
            reactionGroups {content createdAt}
            # reactions(first: 0) {
            #   totalCount
            #   nodes {
            #     # id
            #     user {login avatarUrl url}
            #     content
            #     createdAt
            #   }
            # }
          }
        }
      }
    }
  }
  rateLimit {
    limit
    cost
    remaining
    resetAt
  }
}
"""

COMMENT_REACTIONS_QUERY = """
node_%s: node(id:"%s") {
     ... on IssueComment {
      id
      reactions(first: 100) {
        totalCount
        nodes {
            user {login avatarUrl url}
            content
            createdAt
          }
      }
    }
  }
"""

STARGAZERS_QUERY = """
query($owner: String!, $name: String!, $items_per_page: Int!, $page_after: String) {
  repository(owner: $owner, name: $name) {
    stargazers(first: $items_per_page, orderBy: {field: STARRED_AT, direction: DESC}, after: $page_after) {
      pageInfo {
        endCursor
        startCursor
      }
      edges {
        starredAt
        node {
          login
          avatarUrl
          url
        }
      }
    }
  }
  rateLimit {
    limit
    cost
    remaining
    resetAt
  }
}
"""



from typing import Any, Dict, Iterable, Iterator

import dlt
import pendulum

from .helpers import (
    _get_account,
    _get_campaign_budget,
    _get_campaign_details,
    _get_campaign_jobs,
    _get_campaign_properties,
    _get_campaign_stats,
    _get_oauth_token,
    _get_traffic_report,
    _paginate_campaigns,
)


@dlt.source(name="indeed", max_table_nesting=0)
def indeed_source(
    client_id: str,
    client_secret: str,
    employer_id: str,
    start_date: pendulum.DateTime,
    end_date: pendulum.DateTime | None = None,
) -> Iterable[dlt.sources.DltResource]:
    token = _get_oauth_token(client_id, client_secret, employer_id)

    @dlt.resource(name="campaigns", write_disposition="replace")
    def campaigns() -> Iterator[Dict[str, Any]]:
        for campaign in _paginate_campaigns(token):
            yield campaign

    @dlt.transformer(
        name="campaign_details",
        write_disposition="replace",
        data_from=campaigns,
        parallelized=True,
    )
    def campaign_details(campaign: Dict[str, Any]) -> Iterator[Dict[str, Any]]:
        details = _get_campaign_details(token, campaign["Id"])
        yield details

    @dlt.transformer(
        name="campaign_budget",
        write_disposition="replace",
        data_from=campaigns,
        parallelized=True,
    )
    def campaign_budget(campaign: Dict[str, Any]) -> Iterator[Dict[str, Any]]:
        budget = _get_campaign_budget(token, campaign["Id"])
        if budget:
            yield budget

    @dlt.transformer(
        name="campaign_jobs",
        write_disposition="replace",
        data_from=campaigns,
        parallelized=True,
    )
    def campaign_jobs(campaign: Dict[str, Any]) -> Iterator[Dict[str, Any]]:
        for job in _get_campaign_jobs(token, campaign["Id"]):
            yield job

    @dlt.transformer(
        name="campaign_properties",
        write_disposition="replace",
        data_from=campaigns,
        parallelized=True,
    )
    def campaign_properties(campaign: Dict[str, Any]) -> Iterator[Dict[str, Any]]:
        props = _get_campaign_properties(token, campaign["Id"])
        if props:
            yield props

    @dlt.transformer(
        name="campaign_stats",
        write_disposition="merge",
        merge_key="Date",
        data_from=campaigns,
        parallelized=True,
    )
    def campaign_stats(
        campaign: Dict[str, Any],
        date: dlt.sources.incremental[str] = dlt.sources.incremental(
            "Date",
            initial_value=start_date.to_date_string(),
            end_value=end_date.to_date_string() if end_date else None,
            range_start="closed",
            range_end="closed",
        ),
    ) -> Iterator[Dict[str, Any]]:
        current_start = date.last_value or start_date.to_date_string()
        current_end = date.end_value or pendulum.now("UTC").to_date_string()

        for stat in _get_campaign_stats(
            token, campaign["Id"], current_start, current_end
        ):
            yield stat

    @dlt.resource(
        name="account",
        write_disposition="replace",
    )
    def account() -> Iterator[Dict[str, Any]]:
        data = _get_account(token)
        employer_id = data.get("employerId")
        contact = data.get("contact")
        company = data.get("company")
        email = data.get("email")

        for job_source in data.get("jobSourceList", []):
            yield {
                "employerId": employer_id,
                "contact": contact,
                "company": company,
                "email": email,
                "jobSourceId": job_source.get("id"),
                "jobSourceSiteName": job_source.get("siteName"),
            }

    @dlt.resource(name="traffic_stats", write_disposition="merge", merge_key="date")
    def traffic_stats(
        date: dlt.sources.incremental[str] = dlt.sources.incremental(
            "date",
            initial_value=start_date.to_date_string(),
            end_value=end_date.to_date_string() if end_date else None,
            range_start="closed",
            range_end="closed",
        ),
    ) -> Iterator[Dict[str, Any]]:
        current_start = date.last_value or start_date.to_date_string()
        current_end = date.end_value or pendulum.now("UTC").to_date_string()

        for row in _get_traffic_report(token, current_start, current_end):
            yield row

    return [
        campaigns,
        campaign_details,
        campaign_budget,
        campaign_jobs,
        campaign_properties,
        campaign_stats,
        account,
        traffic_stats,
    ]



from typing import Any

import requests

BASE_URLS = {
    "us": "https://api.customer.io",
    "eu": "https://api-eu.customer.io",
}

# Maximum steps for each metrics period type
MAX_STEPS_BY_PERIOD = {
    "hours": 24,
    "days": 45,
    "weeks": 12,
    "months": 12,
}

# For newsletter metrics, months has a different max
MAX_STEPS_NEWSLETTER = {
    "hours": 24,
    "days": 45,
    "weeks": 12,
    "months": 121,
}


class CustomerIoClient:
    def __init__(self, api_key: str, region: str = "us"):
        self.api_key = api_key
        self.base_url = BASE_URLS.get(region.lower(), BASE_URLS["us"])

    def _get_headers(self):
        return {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }

    def _extract_metrics(
        self,
        result: dict,
        period: str,
        extra_fields: dict | None = None,
    ) -> list:
        """Extract metrics from API response and add common fields."""
        metrics = []
        for i, metric in enumerate(result.get("series", {}).get("series", [])):
            metric["period"] = period
            metric["step_index"] = i
            if extra_fields:
                metric.update(extra_fields)
            metrics.append(metric)
        return metrics

    def _fetch_pages(
        self,
        session: requests.Session,
        url: str,
        params: dict[str, Any] | None = None,
        data_key: str = "activities",
    ) -> list:
        all_items: list[Any] = []
        if params is None:
            params = {}

        while True:
            response = session.get(url=url, headers=self._get_headers(), params=params)
            response.raise_for_status()
            result = response.json()

            items = result.get(data_key) or []
            all_items.extend(items)

            next_token = result.get("next")
            if not next_token:
                break

            params["start"] = next_token

        return all_items

    def fetch_activities(
        self,
        session: requests.Session,
        activity_type: str | None = None,
        name: str | None = None,
        customer_id: str | None = None,
        id_type: str | None = None,
        deleted: bool = False,
        limit: int = 100,
    ):
        url = f"{self.base_url}/v1/activities"
        params = {"limit": limit, "deleted": str(deleted).lower()}

        if activity_type:
            params["type"] = activity_type
        if name:
            params["name"] = name
        if customer_id:
            params["customer_id"] = customer_id
        if id_type:
            params["id_type"] = id_type

        return self._fetch_pages(session, url, params, data_key="activities")

    def fetch_broadcasts(self, session: requests.Session):
        url = f"{self.base_url}/v1/broadcasts"
        return self._fetch_pages(session, url, data_key="broadcasts")

    def fetch_campaigns(self, session: requests.Session):
        url = f"{self.base_url}/v1/campaigns"
        return self._fetch_pages(session, url, data_key="campaigns")

    def fetch_collections(self, session: requests.Session):
        url = f"{self.base_url}/v1/collections"
        return self._fetch_pages(session, url, data_key="collections")

    def fetch_exports(self, session: requests.Session):
        url = f"{self.base_url}/v1/exports"
        return self._fetch_pages(session, url, data_key="exports")

    def fetch_info_ip_addresses(self, session: requests.Session):
        url = f"{self.base_url}/v1/info/ip_addresses"
        response = session.get(url=url, headers=self._get_headers())
        response.raise_for_status()
        result = response.json()
        ips = result.get("ips", [])
        return [{"ip": ip} for ip in ips]

    def fetch_messages(
        self,
        session: requests.Session,
        start_ts: int | None = None,
        end_ts: int | None = None,
    ) -> list:
        url = f"{self.base_url}/v1/messages"
        params = {"limit": 1000}

        if start_ts:
            params["start_ts"] = start_ts
        if end_ts:
            params["end_ts"] = end_ts

        return self._fetch_pages(session, url, params, data_key="messages")

    def fetch_newsletters(self, session: requests.Session) -> list:
        url = f"{self.base_url}/v1/newsletters"
        params = {"limit": 100}
        return self._fetch_pages(session, url, params, data_key="newsletters")

    def fetch_reporting_webhooks(self, session: requests.Session) -> list:
        url = f"{self.base_url}/v1/reporting_webhooks"
        response = session.get(url=url, headers=self._get_headers())
        response.raise_for_status()
        result = response.json()
        return result.get("reporting_webhooks") or []

    def fetch_segments(self, session: requests.Session) -> list:
        url = f"{self.base_url}/v1/segments"
        return self._fetch_pages(session, url, data_key="segments")

    def fetch_transactional_messages(self, session: requests.Session) -> list:
        url = f"{self.base_url}/v1/transactional"
        return self._fetch_pages(session, url, data_key="transactional")

    def fetch_workspaces(self, session: requests.Session) -> list:
        url = f"{self.base_url}/v1/workspaces"
        response = session.get(url=url, headers=self._get_headers())
        response.raise_for_status()
        result = response.json()
        return result.get("workspaces", [])

    def fetch_newsletter_test_groups(
        self, session: requests.Session, newsletter_id: int
    ) -> list:
        url = f"{self.base_url}/v1/newsletters/{newsletter_id}/test_groups"
        response = session.get(url=url, headers=self._get_headers())
        response.raise_for_status()
        result = response.json()

        groups = []
        for group in result.get("test_groups", []):
            group["newsletter_id"] = newsletter_id
            groups.append(group)

        return groups

    def fetch_newsletter_metrics(
        self,
        session: requests.Session,
        newsletter_id: int,
        period: str = "days",
    ) -> list:
        steps = MAX_STEPS_NEWSLETTER.get(period, 45)
        url = f"{self.base_url}/v1/newsletters/{newsletter_id}/metrics"
        params: dict[str, Any] = {"period": period, "steps": steps}

        response = session.get(url=url, headers=self._get_headers(), params=params)
        response.raise_for_status()
        return self._extract_metrics(
            response.json(), period, {"newsletter_id": newsletter_id}
        )

    def fetch_broadcast_metrics(
        self,
        session: requests.Session,
        period: str = "days",
    ) -> list:
        steps = MAX_STEPS_BY_PERIOD.get(period, 45)
        broadcasts = self.fetch_broadcasts(session)
        all_metrics = []

        for broadcast in broadcasts:
            broadcast_id = broadcast.get("id")
            url = f"{self.base_url}/v1/broadcasts/{broadcast_id}/metrics"
            params: dict[str, Any] = {"period": period, "steps": steps}

            response = session.get(url=url, headers=self._get_headers(), params=params)
            response.raise_for_status()
            all_metrics.extend(
                self._extract_metrics(
                    response.json(), period, {"broadcast_id": broadcast_id}
                )
            )

        return all_metrics

    def fetch_broadcast_actions(
        self, session: requests.Session, broadcast_id: int
    ) -> list:
        url = f"{self.base_url}/v1/broadcasts/{broadcast_id}/actions"

        response = session.get(url=url, headers=self._get_headers())
        response.raise_for_status()
        result = response.json()

        actions = []
        for action in result.get("actions", []):
            action["broadcast_id"] = broadcast_id
            actions.append(action)

        return actions

    def fetch_broadcast_messages(
        self,
        session: requests.Session,
        broadcast_id: int,
        start_ts: int | None = None,
        end_ts: int | None = None,
    ) -> list:
        url = f"{self.base_url}/v1/broadcasts/{broadcast_id}/messages"
        params: dict[str, Any] = {}

        if start_ts:
            params["start_ts"] = start_ts
        if end_ts:
            params["end_ts"] = end_ts

        return self._fetch_pages(session, url, params, data_key="messages")

    def fetch_broadcast_action_metrics(
        self,
        session: requests.Session,
        broadcast_id: int,
        action_id: int,
        period: str = "days",
    ) -> list:
        steps = MAX_STEPS_BY_PERIOD.get(period, 45)
        url = (
            f"{self.base_url}/v1/broadcasts/{broadcast_id}/actions/{action_id}/metrics"
        )
        params: dict[str, Any] = {"period": period, "steps": steps}

        response = session.get(url=url, headers=self._get_headers(), params=params)
        response.raise_for_status()
        return self._extract_metrics(
            response.json(),
            period,
            {"broadcast_id": broadcast_id, "action_id": action_id},
        )

    def fetch_campaign_metrics(
        self,
        session: requests.Session,
        campaign_id: int,
        period: str = "days",
        start_ts: int | None = None,
        end_ts: int | None = None,
    ) -> list:
        url = f"{self.base_url}/v1/campaigns/{campaign_id}/metrics"
        params: dict[str, Any] = {"version": 2, "res": period}

        if start_ts:
            params["start"] = start_ts
        if end_ts:
            params["end"] = end_ts

        response = session.get(url=url, headers=self._get_headers(), params=params)
        response.raise_for_status()
        return self._extract_metrics(
            response.json(), period, {"campaign_id": campaign_id}
        )

    def fetch_campaign_actions(
        self, session: requests.Session, campaign_id: int
    ) -> list:
        url = f"{self.base_url}/v1/campaigns/{campaign_id}/actions"
        actions = self._fetch_pages(session, url, data_key="actions")
        for action in actions:
            action["campaign_id"] = campaign_id
        return actions

    def fetch_sender_identities(self, session: requests.Session) -> list:
        url = f"{self.base_url}/v1/sender_identities"
        return self._fetch_pages(session, url, data_key="sender_identities")

    def fetch_campaign_action_metrics(
        self,
        session: requests.Session,
        campaign_id: int,
        action_id: int,
        period: str = "days",
        start_ts: int | None = None,
        end_ts: int | None = None,
    ) -> list:
        url = f"{self.base_url}/v1/campaigns/{campaign_id}/actions/{action_id}/metrics"
        params: dict[str, Any] = {"version": 2, "res": period}

        if start_ts:
            params["start"] = start_ts
        if end_ts:
            params["end"] = end_ts

        response = session.get(url=url, headers=self._get_headers(), params=params)
        response.raise_for_status()
        return self._extract_metrics(
            response.json(),
            period,
            {"campaign_id": campaign_id, "action_id": action_id},
        )

    def fetch_customers(
        self,
        session: requests.Session,
        limit: int = 1000,
    ) -> list:
        """Fetch customers by iterating through all segments and getting their members."""
        # The POST /v1/customers endpoint requires a filter, so we fetch segment members instead
        all_customers = {}
        segments = self.fetch_segments(session)

        for segment in segments:
            segment_id = segment.get("id")
            if segment_id:
                members = self.fetch_segment_members(session, segment_id, limit)
                for member in members:
                    # Use cio_id as unique key to deduplicate across segments
                    cio_id = member.get("cio_id")
                    if cio_id and cio_id not in all_customers:
                        all_customers[cio_id] = member

        return list(all_customers.values())

    def fetch_segment_members(
        self,
        session: requests.Session,
        segment_id: int,
        limit: int = 1000,
    ) -> list:
        """Fetch members of a specific segment."""
        url = f"{self.base_url}/v1/segments/{segment_id}/membership"
        params = {"limit": limit}
        return self._fetch_pages(session, url, params, data_key="identifiers")

    def _fetch_pages_post(
        self,
        session: requests.Session,
        url: str,
        params: dict[str, Any] | None = None,
        data_key: str = "identifiers",
    ) -> list:
        """Pagination helper for POST requests."""
        all_items: list[Any] = []
        if params is None:
            params = {}
        query_params: dict[str, Any] = {}

        while True:
            response = session.post(
                url=url, headers=self._get_headers(), json=params, params=query_params
            )
            response.raise_for_status()
            result = response.json()

            items = result.get(data_key) or []
            all_items.extend(items)

            next_token = result.get("next")
            if not next_token:
                break

            query_params["start"] = next_token

        return all_items

    def fetch_customer_attributes(
        self, session: requests.Session, customer_id: str
    ) -> dict | None:
        """Fetch attributes for a specific customer."""
        url = f"{self.base_url}/v1/customers/{customer_id}/attributes"
        response = session.get(url=url, headers=self._get_headers())
        if response.status_code == 404:
            return None
        response.raise_for_status()
        result = response.json()
        customer = result.get("customer", {})
        customer["customer_id"] = customer_id
        return customer

    def fetch_customer_messages(
        self,
        session: requests.Session,
        customer_id: str,
        start_ts: int | None = None,
        end_ts: int | None = None,
    ) -> list:
        """Fetch messages sent to a specific customer."""
        url = f"{self.base_url}/v1/customers/{customer_id}/messages"
        params = {}
        if start_ts:
            params["start_ts"] = start_ts
        if end_ts:
            params["end_ts"] = end_ts

        messages = self._fetch_pages(session, url, params, data_key="messages")
        for msg in messages:
            msg["customer_id"] = customer_id
        return messages

    def fetch_customer_activities(
        self, session: requests.Session, customer_id: str
    ) -> list:
        """Fetch activities for a specific customer."""
        url = f"{self.base_url}/v1/customers/{customer_id}/activities"
        activities = self._fetch_pages(session, url, data_key="activities")
        for activity in activities:
            activity["customer_id"] = customer_id
        return activities

    def fetch_customer_relationships(
        self, session: requests.Session, customer_id: str
    ) -> list:
        """Fetch object relationships for a specific customer."""
        url = f"{self.base_url}/v1/customers/{customer_id}/relationships"
        relationships = self._fetch_pages(session, url, data_key="cio_relationships")
        for rel in relationships:
            rel["customer_id"] = customer_id
        return relationships

    def fetch_object_types(self, session: requests.Session) -> list:
        """Fetch all object types in the workspace."""
        url = f"{self.base_url}/v1/object_types"
        response = session.get(url=url, headers=self._get_headers())
        response.raise_for_status()
        result = response.json()
        return result.get("types", [])

    def fetch_objects(
        self, session: requests.Session, object_type_id: str, limit: int = 1000
    ) -> list:
        """Fetch objects of a specific type."""
        url = f"{self.base_url}/v1/objects"
        params = {"object_type_id": object_type_id, "limit": limit}
        objects = self._fetch_pages_post(session, url, params, data_key="identifiers")
        for obj in objects:
            obj["object_type_id"] = object_type_id
        return objects

    def fetch_object_attributes(
        self, session: requests.Session, object_type_id: str, object_id: str
    ) -> dict | None:
        """Fetch attributes for a specific object."""
        url = f"{self.base_url}/v1/objects/{object_type_id}/{object_id}/attributes"
        response = session.get(url=url, headers=self._get_headers())
        if response.status_code == 404:
            return None
        response.raise_for_status()
        result = response.json()
        obj = result.get("object", {})
        obj["object_type_id"] = object_type_id
        obj["object_id"] = object_id
        return obj

    def fetch_object_relationships(
        self, session: requests.Session, object_type_id: str, object_id: str
    ) -> list:
        """Fetch people related to a specific object."""
        url = f"{self.base_url}/v1/objects/{object_type_id}/{object_id}/relationships"
        relationships = self._fetch_pages(session, url, data_key="cio_relationships")
        for rel in relationships:
            rel["object_type_id"] = object_type_id
            rel["object_id"] = object_id
        return relationships

    def fetch_campaign_messages(
        self,
        session: requests.Session,
        campaign_id: int,
        start_ts: int | None = None,
        end_ts: int | None = None,
    ) -> list:
        """Fetch messages/deliveries for a specific campaign."""
        url = f"{self.base_url}/v1/campaigns/{campaign_id}/messages"
        params = {}
        if start_ts:
            params["start_ts"] = start_ts
        if end_ts:
            params["end_ts"] = end_ts

        messages = self._fetch_pages(session, url, params, data_key="messages")
        for msg in messages:
            msg["campaign_id"] = campaign_id
        return messages

    def fetch_subscription_topics(self, session: requests.Session) -> list:
        """Fetch subscription topics in the workspace."""
        url = f"{self.base_url}/v1/subscription_topics"
        response = session.get(url=url, headers=self._get_headers())
        response.raise_for_status()
        result = response.json()
        return result.get("topics", [])



from typing import Iterable

import pendulum
from dlt.sources.helpers.requests import Client


class WiseClient:
    BASE_URL = "https://api.transferwise.com"

    def __init__(self, api_key: str) -> None:
        self.session = Client(raise_for_status=False).session
        self.session.headers.update({"Authorization": f"Bearer {api_key}"})

    # https://docs.wise.com/api-docs/api-reference/profile#list-profiles
    def fetch_profiles(self) -> Iterable[dict]:
        url = f"{self.BASE_URL}/v2/profiles"
        resp = self.session.get(url)
        resp.raise_for_status()
        for profile in resp.json():
            yield profile

    # https://docs.wise.com/api-docs/api-reference/transfer#list-transfers
    def fetch_transfers(
        self, profile_id: str, start_time=pendulum.DateTime, end_time=pendulum.DateTime
    ):
        offset = 0

        while True:
            data = self.session.get(
                f"{self.BASE_URL}/v1/transfers",
                params={
                    "profile": profile_id,
                    "createdDateStart": start_time.to_date_string(),
                    "createdDateEnd": end_time.to_date_string(),
                    "limit": 100,
                    "offset": offset,
                },
            )
            response_data = data.json()

            if not response_data or len(response_data) == 0:
                break

            for transfer in response_data:
                transfer["created"] = pendulum.parse(transfer["created"])

                yield transfer
            offset += 100

    # https://docs.wise.com/api-docs/api-reference/balance#list
    def fetch_balances(
        self, profile_id: str, start_time=pendulum.DateTime, end_time=pendulum.DateTime
    ) -> Iterable[dict]:
        url = f"{self.BASE_URL}/v4/profiles/{profile_id}/balances"
        resp = self.session.get(url, params={"types": "STANDARD,SAVINGS"})
        resp.raise_for_status()
        for balance in resp.json():
            balance["modificationTime"] = pendulum.parse(balance["modificationTime"])
            if (
                balance["modificationTime"] > start_time
                and balance["modificationTime"] < end_time
            ):
                yield balance



# Copyright 2022-2025 ScaleVector
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A source that extracts data from Notion API"""

from typing import Dict, Iterator, List, Optional

import dlt
from dlt.sources import DltResource

from .helpers.client import NotionClient
from .helpers.database import NotionDatabase


@dlt.source(max_table_nesting=1)
def notion_databases(
    database_ids: Optional[List[Dict[str, str]]] = None,
    api_key: str = dlt.secrets.value,
) -> Iterator[DltResource]:
    """
    Retrieves data from Notion databases.

    Args:
        database_ids (List[Dict[str, str]], optional): A list of dictionaries
            each containing a database id and a name.
            Defaults to None. If None, the function will generate all databases
            in the workspace that are accessible to the integration.
        api_key (str): The Notion API secret key.

    Yields:
        DltResource: Data resources from Notion databases.
    """
    notion_client = NotionClient(api_key)

    if database_ids is None:
        search_results = notion_client.search(
            filter_criteria={"value": "database", "property": "object"}
        )
        database_ids = [
            {"id": result["id"], "use_name": result["title"][0]["plain_text"]}
            for result in search_results
        ]

    for database in database_ids:
        if "use_name" not in database:
            # Fetch the database details from Notion
            details = notion_client.get_database(database["id"])

            # Extract the name/title from the details
            database["use_name"] = details["title"][0]["plain_text"]

        notion_database = NotionDatabase(database["id"], notion_client)
        yield dlt.resource(  # type: ignore
            notion_database.query(),
            primary_key="id",
            name=database["use_name"],
            write_disposition="replace",
        )



from typing import Iterable, Optional

import dlt
import pendulum
import requests
from dlt.common.typing import TAnyDateTime, TDataItem
from dlt.sources import DltResource
from dlt.sources.helpers.requests import Client

from ingestr.src.phantombuster.client import PhantombusterClient


def retry_on_limit(
    response: Optional[requests.Response], exception: Optional[BaseException]
) -> bool:
    if response is not None and response.status_code == 429:
        return True
    return False


def create_client() -> requests.Session:
    return Client(
        raise_for_status=False,
        retry_condition=retry_on_limit,
        request_max_attempts=12,
        request_backoff_factor=2,
    ).session


@dlt.source(max_table_nesting=0)
def phantombuster_source(
    api_key: str, agent_id: str, start_date: TAnyDateTime, end_date: TAnyDateTime | None
) -> Iterable[DltResource]:
    client = PhantombusterClient(api_key)

    @dlt.resource(
        write_disposition="merge",
        primary_key="container_id",
        columns={
            "partition_dt": {"data_type": "date", "partition": True},
        },
    )
    def completed_phantoms(
        dateTime=(
            dlt.sources.incremental(
                "ended_at",
                initial_value=start_date,
                end_value=end_date,
                range_start="closed",
                range_end="closed",
            )
        ),
    ) -> Iterable[TDataItem]:
        if dateTime.end_value is None:
            end_dt = pendulum.now(tz="UTC")
        else:
            end_dt = dateTime.end_value

        start_dt = dateTime.last_value

        yield client.fetch_containers_result(
            create_client(), agent_id, start_date=start_dt, end_date=end_dt
        )

    return completed_phantoms



# Copyright 2022-2025 ScaleVector
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Shopify source helpers"""

from typing import Any, Iterable, Literal, Optional
from urllib.parse import urljoin

from dlt.common import jsonpath
from dlt.common.time import ensure_pendulum_datetime
from dlt.common.typing import Dict, DictStrAny, TDataItems
from dlt.sources.helpers import requests

from .exceptions import ShopifyPartnerApiError
from .settings import DEFAULT_API_VERSION, DEFAULT_PARTNER_API_VERSION

TOrderStatus = Literal["open", "closed", "cancelled", "any"]


def convert_datetime_fields(item: Dict[str, Any]) -> Dict[str, Any]:
    """Convert timestamp fields in the item to pendulum datetime objects

    The item is modified in place, including nested items.

    Args:
        item: The item to convert

    Returns:
        The same data item (for convenience)
    """
    fields = ["created_at", "updated_at", "createdAt", "updatedAt"]

    def convert_nested(obj: Any) -> Any:
        if isinstance(obj, dict):
            for key, value in obj.items():
                if key in fields and isinstance(value, str):
                    obj[key] = ensure_pendulum_datetime(value)
                else:
                    obj[key] = convert_nested(value)
        elif isinstance(obj, list):
            return [convert_nested(elem) for elem in obj]
        return obj

    return convert_nested(item)


def remove_nodes_key(item: Any) -> Any:
    """
    Recursively remove the 'nodes' key from dictionaries if it's the only key and its value is an array.

    Args:
        item: The item to process (can be a dict, list, or any other type)

    Returns:
        The processed item
    """
    if isinstance(item, dict):
        if len(item) == 1 and "nodes" in item and isinstance(item["nodes"], list):
            return [remove_nodes_key(node) for node in item["nodes"]]
        return {k: remove_nodes_key(v) for k, v in item.items()}
    elif isinstance(item, list):
        return [remove_nodes_key(element) for element in item]
    else:
        return item


class ShopifyApi:
    """
    A Shopify API client that can be used to get pages of data from Shopify.
    """

    def __init__(
        self,
        shop_url: str,
        private_app_password: str,
        api_version: str = DEFAULT_API_VERSION,
    ) -> None:
        """
        Args:
            shop_url: The URL of your shop (e.g. https://my-shop.myshopify.com).
            private_app_password: The private app password to the app on your shop.
            api_version: The API version to use (e.g. 2023-01)
        """
        self.shop_url = shop_url
        self.private_app_password = private_app_password
        self.api_version = api_version

    def get_pages(
        self, resource: str, params: Optional[Dict[str, Any]] = None
    ) -> Iterable[TDataItems]:
        """Get all pages from shopify using requests.
        Iterates through all pages and yield each page items.

        Args:
            resource: The resource to get pages for (e.g. products, orders, customers).
            params: Query params to include in the request.

        Yields:
            List of data items from the page
        """
        url = urljoin(self.shop_url, f"/admin/api/{self.api_version}/{resource}.json")

        resource_last = resource.split("/")[-1]

        headers = {"X-Shopify-Access-Token": self.private_app_password}
        while url:
            response = requests.get(url, params=params, headers=headers)
            response.raise_for_status()
            json = response.json()
            yield [convert_datetime_fields(item) for item in json[resource_last]]
            url = response.links.get("next", {}).get("url")
            # Query params are included in subsequent page URLs
            params = None


class ShopifyGraphQLApi:
    """Client for Shopify GraphQL API"""

    def __init__(
        self,
        access_token: str,
        api_version: str = DEFAULT_PARTNER_API_VERSION,
        base_url: str = "partners.shopify.com",
    ) -> None:
        self.access_token = access_token
        self.api_version = api_version
        self.base_url = base_url

    @property
    def graphql_url(self) -> str:
        if self.base_url.startswith("https://"):
            return f"{self.base_url}/admin/api/{self.api_version}/graphql.json"

        return f"https://{self.base_url}/admin/api/{self.api_version}/graphql.json"

    def run_graphql_query(
        self, query: str, variables: Optional[DictStrAny] = None
    ) -> DictStrAny:
        """Run a graphql query against the Shopify Partner API

        Args:
            query: The query to run
            variables: The variables to include in the query

        Returns:
            The response JSON
        """
        headers = {"X-Shopify-Access-Token": self.access_token}
        response = requests.post(
            self.graphql_url,
            json={"query": query, "variables": variables},
            headers=headers,
        )
        data = response.json()
        if data.get("errors"):
            raise ShopifyPartnerApiError(response.text)
        return data  # type: ignore[no-any-return]

    def get_graphql_pages(
        self,
        query: str,
        data_items_path: jsonpath.TJsonPath,
        pagination_cursor_path: jsonpath.TJsonPath,
        pagination_variable_name: str,
        pagination_cursor_has_next_page_path: Optional[jsonpath.TJsonPath] = None,
        variables: Optional[DictStrAny] = None,
    ) -> Iterable[TDataItems]:
        variables = dict(variables or {})
        while True:
            data = self.run_graphql_query(query, variables)
            data_items = jsonpath.find_values(data_items_path, data)

            if not data_items:
                break

            yield [
                remove_nodes_key(convert_datetime_fields(item)) for item in data_items
            ]

            cursors = jsonpath.find_values(pagination_cursor_path, data)
            if not cursors:
                break

            if pagination_cursor_has_next_page_path:
                has_next_page = jsonpath.find_values(
                    pagination_cursor_has_next_page_path, data
                )
                if not has_next_page or not has_next_page[0]:
                    break

            variables[pagination_variable_name] = cursors[-1]



from typing import Iterable

import dlt
import pendulum
import requests
from dlt.common.typing import TAnyDateTime, TDataItem
from dlt.sources import DltResource
from dlt.sources.helpers.requests import Client

from ingestr.src.customer_io.helpers import CustomerIoClient


def retry_on_limit(
    response: requests.Response | None, exception: BaseException | None
) -> bool:
    if response is None:
        return False
    return response.status_code == 429


def create_client() -> requests.Session:
    return Client(
        raise_for_status=False,
        retry_condition=retry_on_limit,
        request_max_attempts=12,
        request_backoff_factor=2,
    ).session


def _get_incremental_bounds(incremental_state) -> tuple:
    """Extract last_value and end_value from incremental state as pendulum datetimes."""
    last_value = (
        pendulum.instance(incremental_state.last_value).in_tz("UTC")
        if incremental_state.last_value
        else pendulum.datetime(1970, 1, 1, tz="UTC")
    )
    end_value = (
        pendulum.instance(incremental_state.end_value).in_tz("UTC")
        if incremental_state.end_value
        else None
    )
    return last_value, end_value


def _filter_by_timestamp(
    items: Iterable, timestamp_field: str, last_value, end_value
) -> Iterable[TDataItem]:
    """Filter items by timestamp field within the given bounds."""
    for item in items:
        item_updated = pendulum.from_timestamp(item.get(timestamp_field, 0))
        if item_updated >= last_value:
            if end_value is None or item_updated <= end_value:
                item[timestamp_field] = item_updated
                yield item


def _to_timestamp(dt: TAnyDateTime | None) -> int | None:
    """Convert datetime to unix timestamp, or return None."""
    if dt is None:
        return None
    if hasattr(dt, "timestamp"):
        return int(dt.timestamp())  # type: ignore[union-attr, attr-defined]
    # Handle other types by converting through pendulum
    return int(pendulum.instance(dt).timestamp())  # type: ignore[arg-type, attr-defined]


@dlt.source(max_table_nesting=0)
def customer_io_source(
    api_key: str,
    region: str = "us",
    start_date: TAnyDateTime | None = None,
    end_date: TAnyDateTime | None = None,
) -> Iterable[DltResource]:
    client = CustomerIoClient(api_key, region)

    # Pre-compute timestamps for reuse
    start_ts = _to_timestamp(start_date)
    end_ts = _to_timestamp(end_date)

    @dlt.resource(write_disposition="replace", primary_key="id")
    def activities() -> Iterable[TDataItem]:
        yield from client.fetch_activities(create_client())

    @dlt.resource(write_disposition="merge", primary_key="id")
    def broadcasts(
        updated=dlt.sources.incremental(
            "updated",
            initial_value=start_date or pendulum.datetime(1970, 1, 1, tz="UTC"),
            end_value=end_date,
        ),
    ) -> Iterable[TDataItem]:
        last_value, end_value_dt = _get_incremental_bounds(updated)
        yield from _filter_by_timestamp(
            client.fetch_broadcasts(create_client()),
            "updated",
            last_value,
            end_value_dt,
        )

    @dlt.transformer(data_from=broadcasts, write_disposition="merge", primary_key="id")
    def broadcast_actions(broadcast: TDataItem) -> Iterable[TDataItem]:
        broadcast_id = broadcast.get("id")
        start_val = (
            pendulum.instance(start_date).in_tz("UTC")  # type: ignore[arg-type, union-attr, attr-defined]
            if start_date
            else pendulum.datetime(1970, 1, 1, tz="UTC")
        )
        end_val = pendulum.instance(end_date).in_tz("UTC") if end_date else None  # type: ignore[arg-type, union-attr, attr-defined]
        yield from _filter_by_timestamp(
            client.fetch_broadcast_actions(create_client(), broadcast_id),
            "updated",
            start_val,
            end_val,
        )

    @dlt.transformer(data_from=broadcasts, write_disposition="merge", primary_key="id")
    def broadcast_messages(broadcast: TDataItem) -> Iterable[TDataItem]:
        broadcast_id = broadcast.get("id")
        yield from client.fetch_broadcast_messages(
            create_client(), broadcast_id, start_ts, end_ts
        )

    @dlt.resource(write_disposition="merge", primary_key="id")
    def campaigns(
        updated=dlt.sources.incremental(
            "updated",
            initial_value=start_date or pendulum.datetime(1970, 1, 1, tz="UTC"),
            end_value=end_date,
        ),
    ) -> Iterable[TDataItem]:
        last_value, end_value_dt = _get_incremental_bounds(updated)
        yield from _filter_by_timestamp(
            client.fetch_campaigns(create_client()), "updated", last_value, end_value_dt
        )

    @dlt.transformer(data_from=campaigns, write_disposition="merge", primary_key="id")
    def campaign_actions(campaign: TDataItem) -> Iterable[TDataItem]:
        campaign_id = campaign.get("id")
        for item in client.fetch_campaign_actions(create_client(), campaign_id):
            item["updated"] = pendulum.from_timestamp(item.get("updated", 0))
            yield item

    @dlt.resource(write_disposition="merge", primary_key="id")
    def collections(
        updated_at=dlt.sources.incremental(
            "updated_at",
            initial_value=start_date or pendulum.datetime(1970, 1, 1, tz="UTC"),
            end_value=end_date,
        ),
    ) -> Iterable[TDataItem]:
        last_value, end_value_dt = _get_incremental_bounds(updated_at)
        yield from _filter_by_timestamp(
            client.fetch_collections(create_client()),
            "updated_at",
            last_value,
            end_value_dt,
        )

    @dlt.resource(write_disposition="merge", primary_key="id")
    def exports(
        updated_at=dlt.sources.incremental(
            "updated_at",
            initial_value=start_date or pendulum.datetime(1970, 1, 1, tz="UTC"),
            end_value=end_date,
        ),
    ) -> Iterable[TDataItem]:
        last_value, end_value_dt = _get_incremental_bounds(updated_at)
        yield from _filter_by_timestamp(
            client.fetch_exports(create_client()),
            "updated_at",
            last_value,
            end_value_dt,
        )

    @dlt.resource(write_disposition="replace", primary_key="ip")
    def info_ip_addresses() -> Iterable[TDataItem]:
        yield from client.fetch_info_ip_addresses(create_client())

    @dlt.resource(write_disposition="merge", primary_key="id")
    def messages() -> Iterable[TDataItem]:
        yield from client.fetch_messages(create_client(), start_ts, end_ts)

    @dlt.resource(write_disposition="merge", primary_key="id")
    def newsletters(
        updated=dlt.sources.incremental(
            "updated",
            initial_value=start_date or pendulum.datetime(1970, 1, 1, tz="UTC"),
            end_value=end_date,
        ),
    ) -> Iterable[TDataItem]:
        last_value, end_value_dt = _get_incremental_bounds(updated)
        yield from _filter_by_timestamp(
            client.fetch_newsletters(create_client()),
            "updated",
            last_value,
            end_value_dt,
        )

    @dlt.transformer(
        data_from=newsletters, write_disposition="replace", primary_key="id"
    )
    def newsletter_test_groups(newsletter: TDataItem) -> Iterable[TDataItem]:
        newsletter_id = newsletter.get("id")
        yield from client.fetch_newsletter_test_groups(create_client(), newsletter_id)

    @dlt.resource(write_disposition="replace", primary_key="id")
    def reporting_webhooks() -> Iterable[TDataItem]:
        yield from client.fetch_reporting_webhooks(create_client())

    @dlt.resource(write_disposition="merge", primary_key="id")
    def segments(
        updated_at=dlt.sources.incremental(
            "updated_at",
            initial_value=start_date or pendulum.datetime(1970, 1, 1, tz="UTC"),
            end_value=end_date,
        ),
    ) -> Iterable[TDataItem]:
        last_value, end_value_dt = _get_incremental_bounds(updated_at)
        yield from _filter_by_timestamp(
            client.fetch_segments(create_client()),
            "updated_at",
            last_value,
            end_value_dt,
        )

    @dlt.resource(write_disposition="replace", primary_key="id")
    def transactional_messages() -> Iterable[TDataItem]:
        yield from client.fetch_transactional_messages(create_client())

    @dlt.resource(write_disposition="replace", primary_key="id")
    def workspaces() -> Iterable[TDataItem]:
        yield from client.fetch_workspaces(create_client())

    @dlt.resource(write_disposition="replace", primary_key="id")
    def sender_identities() -> Iterable[TDataItem]:
        yield from client.fetch_sender_identities(create_client())

    @dlt.resource(write_disposition="replace", primary_key="cio_id")
    def customers() -> Iterable[TDataItem]:
        yield from client.fetch_customers(create_client())

    @dlt.transformer(
        data_from=customers, write_disposition="replace", primary_key="customer_id"
    )
    def customer_attributes(customer: TDataItem) -> Iterable[TDataItem]:
        customer_id = customer.get("cio_id") or customer.get("id")
        if customer_id:
            result = client.fetch_customer_attributes(create_client(), customer_id)
            if result:
                yield result

    @dlt.transformer(data_from=customers, write_disposition="merge", primary_key="id")
    def customer_messages(customer: TDataItem) -> Iterable[TDataItem]:
        customer_id = customer.get("cio_id") or customer.get("id")
        if customer_id:
            yield from client.fetch_customer_messages(
                create_client(), customer_id, start_ts, end_ts
            )

    @dlt.transformer(data_from=customers, write_disposition="replace", primary_key="id")
    def customer_activities(customer: TDataItem) -> Iterable[TDataItem]:
        customer_id = customer.get("cio_id") or customer.get("id")
        if customer_id:
            yield from client.fetch_customer_activities(create_client(), customer_id)

    @dlt.transformer(
        data_from=customers,
        write_disposition="replace",
        primary_key=["customer_id", "object_type_id", "object_id"],
    )
    def customer_relationships(customer: TDataItem) -> Iterable[TDataItem]:
        customer_id = customer.get("cio_id") or customer.get("id")
        if customer_id:
            for rel in client.fetch_customer_relationships(
                create_client(), customer_id
            ):
                identifiers = rel.get("identifiers", {})
                rel["object_id"] = identifiers.get("object_id") or identifiers.get(
                    "cio_object_id"
                )
                yield rel

    @dlt.resource(write_disposition="replace", primary_key="id")
    def object_types() -> Iterable[TDataItem]:
        yield from client.fetch_object_types(create_client())

    @dlt.transformer(
        data_from=object_types,
        write_disposition="replace",
        primary_key=["object_type_id", "object_id"],
    )
    def objects(obj_type: TDataItem) -> Iterable[TDataItem]:
        object_type_id = obj_type.get("id")
        if object_type_id:
            yield from client.fetch_objects(create_client(), str(object_type_id))

    @dlt.resource(write_disposition="replace", primary_key="id")
    def subscription_topics() -> Iterable[TDataItem]:
        yield from client.fetch_subscription_topics(create_client())

    @dlt.transformer(data_from=campaigns, write_disposition="merge", primary_key="id")
    def campaign_messages(campaign: TDataItem) -> Iterable[TDataItem]:
        campaign_id = campaign.get("id")
        yield from client.fetch_campaign_messages(
            create_client(), campaign_id, start_ts, end_ts
        )

    return (
        activities,
        broadcasts,
        broadcast_actions,
        broadcast_messages,
        campaigns,
        campaign_actions,
        campaign_messages,
        collections,
        exports,
        info_ip_addresses,
        messages,
        newsletters,
        newsletter_test_groups,
        reporting_webhooks,
        segments,
        sender_identities,
        transactional_messages,
        workspaces,
        customers,
        customer_attributes,
        customer_messages,
        customer_activities,
        customer_relationships,
        object_types,
        objects,
        subscription_topics,
    )


@dlt.source(max_table_nesting=0)
def customer_io_broadcast_metrics_source(
    api_key: str,
    region: str = "us",
    period: str = "days",
) -> Iterable[DltResource]:
    client = CustomerIoClient(api_key, region)

    @dlt.resource(write_disposition="replace")
    def broadcast_metrics() -> Iterable[TDataItem]:
        yield from client.fetch_broadcast_metrics(create_client(), period)

    return (broadcast_metrics,)


@dlt.source(max_table_nesting=0)
def customer_io_broadcast_action_metrics_source(
    api_key: str,
    region: str = "us",
    period: str = "days",
) -> Iterable[DltResource]:
    client = CustomerIoClient(api_key, region)

    @dlt.resource(write_disposition="replace", primary_key="id", selected=False)
    def broadcasts() -> Iterable[TDataItem]:
        for item in client.fetch_broadcasts(create_client()):
            yield item

    @dlt.transformer(data_from=broadcasts, write_disposition="replace", selected=False)
    def broadcast_actions(broadcast: TDataItem) -> Iterable[TDataItem]:
        broadcast_id = broadcast.get("id")
        for item in client.fetch_broadcast_actions(create_client(), broadcast_id):
            yield item

    @dlt.transformer(
        data_from=broadcast_actions,
        write_disposition="replace",
        primary_key=["broadcast_id", "action_id", "period", "step_index"],
    )
    def broadcast_action_metrics(action: TDataItem) -> Iterable[TDataItem]:
        broadcast_id = action.get("broadcast_id")
        action_id = action.get("id")
        for item in client.fetch_broadcast_action_metrics(
            create_client(), broadcast_id, action_id, period
        ):
            yield item

    return (broadcasts | broadcast_actions | broadcast_action_metrics,)


@dlt.source(max_table_nesting=0)
def customer_io_campaign_metrics_source(
    api_key: str,
    region: str = "us",
    period: str = "days",
    start_date: TAnyDateTime | None = None,
    end_date: TAnyDateTime | None = None,
) -> Iterable[DltResource]:
    client = CustomerIoClient(api_key, region)

    start_ts = _to_timestamp(start_date)
    end_ts = _to_timestamp(end_date)

    @dlt.resource(write_disposition="replace", primary_key="id", selected=False)
    def campaigns() -> Iterable[TDataItem]:
        for item in client.fetch_campaigns(create_client()):
            yield item

    @dlt.transformer(
        data_from=campaigns,
        write_disposition="replace",
        primary_key=["campaign_id", "period", "step_index"],
    )
    def campaign_metrics(campaign: TDataItem) -> Iterable[TDataItem]:
        campaign_id = campaign.get("id")
        for item in client.fetch_campaign_metrics(
            create_client(), campaign_id, period, start_ts, end_ts
        ):
            yield item

    return (campaigns | campaign_metrics,)


@dlt.source(max_table_nesting=0)
def customer_io_campaign_action_metrics_source(
    api_key: str,
    region: str = "us",
    period: str = "days",
    start_date: TAnyDateTime | None = None,
    end_date: TAnyDateTime | None = None,
) -> Iterable[DltResource]:
    client = CustomerIoClient(api_key, region)

    start_ts = _to_timestamp(start_date)
    end_ts = _to_timestamp(end_date)

    @dlt.resource(write_disposition="replace", primary_key="id", selected=False)
    def campaigns() -> Iterable[TDataItem]:
        for item in client.fetch_campaigns(create_client()):
            yield item

    @dlt.transformer(data_from=campaigns, write_disposition="replace", selected=False)
    def campaign_actions(campaign: TDataItem) -> Iterable[TDataItem]:
        campaign_id = campaign.get("id")
        for item in client.fetch_campaign_actions(create_client(), campaign_id):
            yield item

    @dlt.transformer(
        data_from=campaign_actions,
        write_disposition="replace",
        primary_key=["campaign_id", "action_id", "period", "step_index"],
    )
    def campaign_action_metrics(action: TDataItem) -> Iterable[TDataItem]:
        campaign_id = action.get("campaign_id")
        action_id = action.get("id")
        for item in client.fetch_campaign_action_metrics(
            create_client(), campaign_id, action_id, period, start_ts, end_ts
        ):
            yield item

    return (campaigns | campaign_actions | campaign_action_metrics,)


@dlt.source(max_table_nesting=0)
def customer_io_newsletter_metrics_source(
    api_key: str,
    region: str = "us",
    period: str = "days",
) -> Iterable[DltResource]:
    client = CustomerIoClient(api_key, region)

    @dlt.resource(write_disposition="replace", primary_key="id", selected=False)
    def newsletters() -> Iterable[TDataItem]:
        for item in client.fetch_newsletters(create_client()):
            yield item

    @dlt.transformer(
        data_from=newsletters,
        write_disposition="replace",
        primary_key=["newsletter_id", "period", "step_index"],
    )
    def newsletter_metrics(newsletter: TDataItem) -> Iterable[TDataItem]:
        newsletter_id = newsletter.get("id")
        for item in client.fetch_newsletter_metrics(
            create_client(), newsletter_id, period
        ):
            yield item

    return (newsletters | newsletter_metrics,)



"""Simple Trustpilot API client."""

from typing import Any, Dict, Iterable

import pendulum
from dlt.sources.helpers import requests


class TrustpilotClient:
    """Client for the Trustpilot public API."""

    def __init__(self, api_key: str) -> None:
        self.api_key = api_key
        self.base_url = "https://api.trustpilot.com/v1"

    def _get(self, endpoint: str, params: Dict[str, Any]) -> Dict[str, Any]:
        params = dict(params)
        params["apikey"] = self.api_key
        response = requests.get(f"{self.base_url}{endpoint}", params=params)
        response.raise_for_status()
        return response.json()

    def paginated_reviews(
        self,
        business_unit_id: str,
        updated_since: str,
        end_date: str,
        per_page: int = 1000,
    ) -> Iterable[Dict[str, Any]]:
        page = 1
        while True:
            params: Dict[str, Any] = {"perPage": per_page, "page": page}
            if updated_since:
                params["updatedSince"] = updated_since
            data = self._get(f"/business-units/{business_unit_id}/reviews", params)
            reviews = data.get("reviews", data)
            if not reviews:
                break
            for review in reviews:
                end_date_dt = pendulum.parse(end_date)
                review["updated_at"] = review["updatedAt"]
                review_dt = pendulum.parse(review["updated_at"])
                if review_dt > end_date_dt:  # type: ignore
                    continue
                yield review
            if len(reviews) < per_page:
                break
            page += 1



"""Fundraiseup source for ingesting donations, events, fundraisers, recurring plans, and supporters."""

from typing import Any, Dict, Generator, Iterable, TypedDict

import dlt
import pendulum
from dlt.common.time import ensure_pendulum_datetime
from dlt.sources import DltResource

from .client import FundraiseupClient


class DonationCursor(TypedDict):
    id: str
    created_at: pendulum.DateTime


def order_by_created(record) -> DonationCursor:
    last_value = None
    if len(record) == 1:
        (record,) = record
    else:
        record, last_value = record

    cursor: DonationCursor = {
        "id": record["id"],
        "created_at": ensure_pendulum_datetime(record["created_at"]),
    }

    if last_value is None:
        return cursor

    return max(cursor, last_value, key=lambda v: v["created_at"])


@dlt.source(name="fundraiseup", max_table_nesting=0)
def fundraiseup_source(api_key: str) -> Iterable[DltResource]:
    """
    Return resources for Fundraiseup API.

    Args:
        api_key: API key for authentication

    Returns:
        Iterable of DLT resources
    """
    client = FundraiseupClient(api_key=api_key)

    # Define available resources and their configurations
    resources = {
        "donations": {"write_disposition": "replace", "primary_key": "id"},
        "events": {"write_disposition": "replace", "primary_key": "id"},
        "fundraisers": {"write_disposition": "replace", "primary_key": "id"},
        "recurring_plans": {"write_disposition": "replace", "primary_key": "id"},
        "supporters": {"write_disposition": "replace", "primary_key": "id"},
    }

    def create_resource(resource_name: str, config: Dict[str, Any]) -> DltResource:
        """Create a DLT resource dynamically."""

        @dlt.resource(
            name=resource_name,
            write_disposition=config["write_disposition"],
            primary_key=config["primary_key"],
        )
        def generic_resource() -> Generator[Dict[str, Any], None, None]:
            """Generic resource that yields batches directly."""
            for batch in client.get_paginated_data(resource_name):
                yield batch  # type: ignore[misc]

        return generic_resource()

    @dlt.resource(
        name="donations:incremental",
        write_disposition="merge",
        primary_key="id",
    )
    def donations_incremental(
        last_record: dlt.sources.incremental[DonationCursor] = dlt.sources.incremental(
            "$",
            range_start="closed",
            range_end="closed",
            last_value_func=order_by_created,
        ),
    ):
        params = {}
        if last_record.last_value is not None:
            params["starting_after"] = last_record.last_value["id"]
        for batch in client.get_paginated_data("donations", params=params):
            yield batch  # type: ignore[misc]

    # Return all resources
    return [donations_incremental] + [
        create_resource(name, config) for name, config in resources.items()
    ]



# Copyright 2022-2025 ScaleVector
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This source uses Stripe API and dlt to load data such as Customer, Subscription, Event etc. to the database and to calculate the MRR and churn rate."""

from typing import Any, Dict, Generator, Iterable, Optional, Tuple

import dlt
import stripe
from dlt.sources import DltResource
from pendulum import DateTime

from .helpers import (
    async_parallel_pagination,
    generate_date_ranges,
    pagination,
    transform_date,
)


@dlt.source(max_table_nesting=0)
def stripe_source(
    endpoints: Tuple[str, ...],
    stripe_secret_key: str = dlt.secrets.value,
    start_date: Optional[DateTime] = None,
    end_date: Optional[DateTime] = None,
) -> Iterable[DltResource]:
    """
    Retrieves data from the Stripe API for the specified endpoints.

    For all endpoints, Stripe API responses do not provide the key "updated",
    so in most cases, we are forced to load the data in 'replace' mode.
    This source is suitable for all types of endpoints, including 'Events', 'Invoice', etc.
    but these endpoints can also be loaded in incremental mode (see source incremental_stripe_source).

    Args:
        endpoints (Tuple[str, ...]): A tuple of endpoint names to retrieve data from. Defaults to most popular Stripe API endpoints.
        stripe_secret_key (str): The API access token for authentication. Defaults to the value in the `dlt.secrets` object.
        start_date (Optional[DateTime]): An optional start date to limit the data retrieved. Format: datetime(YYYY, MM, DD). Defaults to None.
        end_date (Optional[DateTime]): An optional end date to limit the data retrieved. Format: datetime(YYYY, MM, DD). Defaults to None.

    Returns:
        Iterable[DltResource]: Resources with data that was created during the period greater than or equal to 'start_date' and less than 'end_date'.
    """
    stripe.api_key = stripe_secret_key
    stripe.api_version = "2022-11-15"

    def stripe_resource(
        endpoint: str,
    ) -> Generator[Dict[Any, Any], Any, None]:
        yield from pagination(endpoint, start_date, end_date)

    for endpoint in endpoints:
        yield dlt.resource(
            stripe_resource,
            name=endpoint,
            write_disposition="replace",
        )(endpoint)


@dlt.source(max_table_nesting=0)
def async_stripe_source(
    endpoints: Tuple[str, ...],
    stripe_secret_key: str = dlt.secrets.value,
    start_date: Optional[DateTime] = None,
    end_date: Optional[DateTime] = None,
    max_workers: int = 4,
    rate_limit_delay: float = 0.03,
) -> Iterable[DltResource]:
    """
    ULTRA-FAST async Stripe source optimized for maximum speed and throughput.

    WARNING: Returns data in RANDOM ORDER for maximum performance.
    Uses aggressive concurrency and minimal delays to maximize API throughput.

    Args:
        endpoints (Tuple[str, ...]): A tuple of endpoint names to retrieve data from.
        stripe_secret_key (str): The API access token for authentication. Defaults to the value in the `dlt.secrets` object.
        start_date (Optional[DateTime]): An optional start date to limit the data retrieved. Format: datetime(YYYY, MM, DD). Defaults to 2010-01-01.
        end_date (Optional[DateTime]): An optional end date to limit the data retrieved. Format: datetime(YYYY, MM, DD). Defaults to today.
        max_workers (int): Maximum number of concurrent async tasks. Defaults to 40 for maximum speed.
        rate_limit_delay (float): Minimal delay between requests. Defaults to 0.03 seconds.

    Returns:
        Iterable[DltResource]: Resources with data in RANDOM ORDER (optimized for speed).
    """
    stripe.api_key = stripe_secret_key
    stripe.api_version = "2022-11-15"

    async def async_stripe_resource(endpoint: str):
        yield async_parallel_pagination(endpoint, max_workers, rate_limit_delay)

    for endpoint in endpoints:
        yield dlt.resource(
            async_stripe_resource,
            name=endpoint,
            write_disposition="replace",
        )(endpoint)


@dlt.source(max_table_nesting=0)
def incremental_stripe_source(
    endpoints: Tuple[str, ...],
    stripe_secret_key: str = dlt.secrets.value,
    initial_start_date: Optional[DateTime] = None,
    end_date: Optional[DateTime] = None,
) -> Iterable[DltResource]:
    stripe.api_key = stripe_secret_key
    stripe.api_version = "2022-11-15"
    start_date_unix = (
        transform_date(initial_start_date) if initial_start_date is not None else -1
    )

    for endpoint in endpoints:

        def date_range_resource(
            endpoint: str = endpoint,
            created: Optional[Any] = dlt.sources.incremental(
                "created",
                initial_value=start_date_unix,
                end_value=transform_date(end_date) if end_date is not None else None,
                range_end="closed",
                range_start="closed",
            ),
        ) -> Generator[Dict[str, Any], None, None]:
            from dlt.common import pendulum

            # Use 2010-01-01 as default start (Stripe founding year) to avoid
            # generating hundreds of thousands of hourly ranges from 1969
            default_start_ts = int(pendulum.datetime(2010, 1, 1).timestamp())
            start_ts = (
                created.last_value
                if created.last_value is not None
                else start_date_unix
            )
            if start_ts < 0:
                start_ts = default_start_ts
            end_ts = (
                created.end_value
                if created.end_value is not None
                else int(pendulum.now().timestamp())
            )
            for date_range in generate_date_ranges(start_ts, end_ts):
                date_range["endpoint"] = endpoint
                date_range["created"] = date_range["end_ts"]
                yield date_range

        def fetch_date_range(
            date_range: Dict[str, int],
        ) -> Generator[Dict[Any, Any], Any, None]:
            """Transformer that fetches data for a given date range."""
            yield from pagination(
                date_range["endpoint"],
                start_date=date_range["start_ts"],
                end_date=date_range["end_ts"],
            )

        date_ranges = dlt.resource(
            date_range_resource,
            name=f"{endpoint}_date_ranges",
        )()

        yield (
            date_ranges
            | dlt.transformer(
                fetch_date_range,
                name=endpoint,
                write_disposition="merge",
                primary_key="id",
                parallelized=True,
            )
        )



from typing import Iterable

import dlt
import pendulum
from dlt.common.typing import TDataItem
from dlt.sources import DltResource

from ingestr.src.appsflyer.client import AppsflyerClient

DIMENSION_RESPONSE_MAPPING = {
    "c": "campaign",
    "af_adset_id": "adset_id",
    "af_adset": "adset",
    "af_ad_id": "ad_id",
}
HINTS = {
    "app_id": {
        "data_type": "text",
        "nullable": False,
    },
    "campaign": {
        "data_type": "text",
        "nullable": False,
    },
    "geo": {
        "data_type": "text",
        "nullable": False,
    },
    "cost": {
        "data_type": "decimal",
        "precision": 30,
        "scale": 5,
        "nullable": False,
    },
    "clicks": {
        "data_type": "bigint",
        "nullable": False,
    },
    "impressions": {
        "data_type": "bigint",
        "nullable": False,
    },
    "average_ecpi": {
        "data_type": "decimal",
        "precision": 30,
        "scale": 5,
        "nullable": False,
    },
    "installs": {
        "data_type": "bigint",
        "nullable": False,
    },
    "retention_day_7": {
        "data_type": "decimal",
        "precision": 30,
        "scale": 5,
        "nullable": False,
    },
    "retention_day_14": {
        "data_type": "decimal",
        "precision": 30,
        "scale": 5,
        "nullable": False,
    },
    "cohort_day_1_revenue_per_user": {
        "data_type": "decimal",
        "precision": 30,
        "scale": 5,
        "nullable": True,
    },
    "cohort_day_1_total_revenue_per_user": {
        "data_type": "decimal",
        "precision": 30,
        "scale": 5,
        "nullable": True,
    },
    "cohort_day_3_revenue_per_user": {
        "data_type": "decimal",
        "precision": 30,
        "scale": 5,
        "nullable": True,
    },
    "cohort_day_3_total_revenue_per_user": {
        "data_type": "decimal",
        "precision": 30,
        "scale": 5,
        "nullable": True,
    },
    "cohort_day_7_revenue_per_user": {
        "data_type": "decimal",
        "precision": 30,
        "scale": 5,
        "nullable": True,
    },
    "cohort_day_7_total_revenue_per_user": {
        "data_type": "decimal",
        "precision": 30,
        "scale": 5,
        "nullable": True,
    },
    "cohort_day_14_revenue_per_user": {
        "data_type": "decimal",
        "precision": 30,
        "scale": 5,
        "nullable": True,
    },
    "cohort_day_14_total_revenue_per_user": {
        "data_type": "decimal",
        "precision": 30,
        "scale": 5,
        "nullable": True,
    },
    "cohort_day_21_revenue_per_user": {
        "data_type": "decimal",
        "precision": 30,
        "scale": 5,
        "nullable": True,
    },
    "cohort_day_21_total_revenue_per_user": {
        "data_type": "decimal",
        "precision": 30,
        "scale": 5,
        "nullable": True,
    },
    "install_time": {
        "data_type": "date",
        "nullable": False,
    },
    "loyal_users": {
        "data_type": "bigint",
        "nullable": False,
    },
    "revenue": {
        "data_type": "decimal",
        "precision": 30,
        "scale": 5,
        "nullable": True,
    },
    "roi": {
        "data_type": "decimal",
        "precision": 30,
        "scale": 5,
        "nullable": True,
    },
    "uninstalls": {
        "data_type": "bigint",
        "nullable": True,
    },
}

CAMPAIGNS_DIMENSIONS = ["c", "geo", "app_id", "install_time"]
CAMPAIGNS_METRICS = [
    "average_ecpi",
    "clicks",
    "cohort_day_1_revenue_per_user",
    "cohort_day_1_total_revenue_per_user",
    "cohort_day_14_revenue_per_user",
    "cohort_day_14_total_revenue_per_user",
    "cohort_day_21_revenue_per_user",
    "cohort_day_21_total_revenue_per_user",
    "cohort_day_3_revenue_per_user",
    "cohort_day_3_total_revenue_per_user",
    "cohort_day_7_revenue_per_user",
    "cohort_day_7_total_revenue_per_user",
    "cost",
    "impressions",
    "installs",
    "loyal_users",
    "retention_day_7",
    "revenue",
    "roi",
    "uninstalls",
]

CREATIVES_DIMENSIONS = [
    "c",
    "geo",
    "app_id",
    "install_time",
    "af_adset_id",
    "af_adset",
    "af_ad_id",
]
CREATIVES_METRICS = [
    "impressions",
    "clicks",
    "installs",
    "cost",
    "revenue",
    "average_ecpi",
    "loyal_users",
    "uninstalls",
    "roi",
]


@dlt.source(max_table_nesting=0)
def appsflyer_source(
    api_key: str,
    start_date: str,
    end_date: str,
    dimensions: list[str],
    metrics: list[str],
) -> Iterable[DltResource]:
    client = AppsflyerClient(api_key)

    @dlt.resource(
        write_disposition="merge",
        merge_key="install_time",
        columns=make_hints(CAMPAIGNS_DIMENSIONS, CAMPAIGNS_METRICS),
    )
    def campaigns(
        datetime=dlt.sources.incremental(
            "install_time",
            initial_value=(
                start_date
                if start_date
                else pendulum.today().subtract(days=30).format("YYYY-MM-DD")
            ),
            end_value=end_date,
            range_end="closed",
            range_start="closed",
        ),
    ) -> Iterable[TDataItem]:
        end = (
            datetime.end_value
            if datetime.end_value
            else pendulum.now().format("YYYY-MM-DD")
        )

        yield from client._fetch_data(
            from_date=datetime.last_value,
            to_date=end,
            dimensions=CAMPAIGNS_DIMENSIONS,
            metrics=CAMPAIGNS_METRICS,
        )

    @dlt.resource(
        write_disposition="merge",
        merge_key="install_time",
        columns=make_hints(CREATIVES_DIMENSIONS, CREATIVES_METRICS),
    )
    def creatives(
        datetime=dlt.sources.incremental(
            "install_time",
            initial_value=(
                start_date
                if start_date
                else pendulum.today().subtract(days=30).format("YYYY-MM-DD")
            ),
            end_value=end_date,
            range_end="closed",
            range_start="closed",
        ),
    ) -> Iterable[TDataItem]:
        end = (
            datetime.end_value
            if datetime.end_value
            else pendulum.now().format("YYYY-MM-DD")
        )
        yield from client._fetch_data(
            datetime.last_value,
            end,
            dimensions=CREATIVES_DIMENSIONS,
            metrics=CREATIVES_METRICS,
        )

    primary_keys = []
    if "install_time" not in dimensions:
        dimensions.append("install_time")
        primary_keys.append("install_time")

    for dimension in dimensions:
        if dimension in DIMENSION_RESPONSE_MAPPING:
            primary_keys.append(DIMENSION_RESPONSE_MAPPING[dimension])
        else:
            primary_keys.append(dimension)

    @dlt.resource(
        write_disposition="merge",
        primary_key=primary_keys,
        columns=make_hints(dimensions, metrics),
    )
    def custom(
        datetime=dlt.sources.incremental(
            "install_time",
            initial_value=(
                start_date
                if start_date
                else pendulum.today().subtract(days=30).format("YYYY-MM-DD")
            ),
            end_value=end_date,
        ),
    ):
        end = (
            datetime.end_value
            if datetime.end_value
            else pendulum.now().format("YYYY-MM-DD")
        )
        res = client._fetch_data(
            from_date=datetime.last_value,
            to_date=end,
            dimensions=dimensions,
            metrics=metrics,
        )
        yield from res

    return campaigns, creatives, custom


def make_hints(dimensions: list[str], metrics: list[str]):
    campaign_hints = {}
    for dimension in dimensions:
        resp_key = dimension
        if dimension in DIMENSION_RESPONSE_MAPPING:
            resp_key = DIMENSION_RESPONSE_MAPPING[dimension]

        if resp_key in HINTS:
            campaign_hints[resp_key] = HINTS[resp_key]

    for metric in metrics:
        if metric in HINTS:
            campaign_hints[metric] = HINTS[metric]

    return campaign_hints



import json
from typing import Iterable

import pendulum
from dlt.sources.helpers.requests import Client


class MixpanelClient:
    def __init__(self, username: str, password: str, project_id: str, server: str):
        self.username = username
        self.password = password
        self.project_id = project_id
        self.server = server
        self.session = Client(raise_for_status=False).session

    def fetch_events(
        self, start_date: pendulum.DateTime, end_date: pendulum.DateTime
    ) -> Iterable[dict]:
        if self.server == "us":
            server = "data"
        elif self.server == "in":
            server = "data-in"
        else:
            server = "data-eu"

        url = f"https://{server}.mixpanel.com/api/2.0/export/"
        params = {
            "project_id": self.project_id,
            "from_date": start_date.format("YYYY-MM-DD"),
            "to_date": end_date.format("YYYY-MM-DD"),
        }
        headers = {
            "accept": "text/plain",
        }
        from requests.auth import HTTPBasicAuth

        auth = HTTPBasicAuth(self.username, self.password)
        resp = self.session.get(url, params=params, headers=headers, auth=auth)
        resp.raise_for_status()
        for line in resp.iter_lines():
            if line:
                data = json.loads(line.decode())
                if "properties" in data:
                    for key, value in data["properties"].items():
                        if key.startswith("$"):
                            data[key[1:]] = value
                        else:
                            data[key] = value
                    del data["properties"]
                yield data

    def fetch_profiles(
        self, start_date: pendulum.DateTime, end_date: pendulum.DateTime
    ) -> Iterable[dict]:
        if self.server == "us":
            server = ""
        elif self.server == "in":
            server = "in."
        else:
            server = "eu."
        url = f"https://{server}mixpanel.com/api/query/engage"
        headers = {
            "accept": "application/json",
            "content-type": "application/x-www-form-urlencoded",
        }
        from requests.auth import HTTPBasicAuth

        auth = HTTPBasicAuth(self.username, self.password)
        page = 0
        session_id = None
        while True:
            params = {"project_id": self.project_id, "page": str(page)}
            if session_id:
                params["session_id"] = session_id
            start_str = start_date.format("YYYY-MM-DDTHH:mm:ss")
            end_str = end_date.format("YYYY-MM-DDTHH:mm:ss")
            where = f'properties["$last_seen"] >= "{start_str}" and properties["$last_seen"] <= "{end_str}"'
            params["where"] = where
            resp = self.session.post(url, params=params, headers=headers, auth=auth)

            resp.raise_for_status()
            data = resp.json()

            for result in data.get("results", []):
                for key, value in result["$properties"].items():
                    if key.startswith("$"):
                        if key == "$last_seen":
                            result["last_seen"] = pendulum.parse(value)
                        else:
                            result[key[1:]] = value
                result["distinct_id"] = result["$distinct_id"]
                del result["$properties"]
                del result["$distinct_id"]
                yield result
            if not data.get("results"):
                break
            session_id = data.get("session_id", session_id)

            page += 1



# Copyright 2022-2025 ScaleVector
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Mongo database source helpers and destination utilities"""

import re
from itertools import islice
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterable,
    Iterator,
    List,
    Mapping,
    Optional,
    Tuple,
    Union,
)

import dlt
from bson.decimal128 import Decimal128
from bson.objectid import ObjectId
from bson.regex import Regex
from bson.timestamp import Timestamp
from dlt.common import logger
from dlt.common.configuration.specs import BaseConfiguration, configspec
from dlt.common.data_writers import TDataItemFormat
from dlt.common.time import ensure_pendulum_datetime
from dlt.common.typing import TDataItem
from dlt.common.utils import map_nested_in_place
from pendulum import _datetime
from pymongo import ASCENDING, DESCENDING, MongoClient
from pymongo.collection import Collection
from pymongo.cursor import Cursor
from pymongo.helpers_shared import _fields_list_to_dict

if TYPE_CHECKING:
    TMongoClient = MongoClient[Any]
    TCollection = Collection[Any]
    TCursor = Cursor[Any]
else:
    TMongoClient = Any
    TCollection = Any
    TCursor = Any

try:
    import pymongoarrow  # type: ignore

    PYMONGOARROW_AVAILABLE = True
except ImportError:
    PYMONGOARROW_AVAILABLE = False


class CollectionLoader:
    def __init__(
        self,
        client: TMongoClient,
        collection: TCollection,
        chunk_size: int,
        incremental: Optional[dlt.sources.incremental[Any]] = None,
    ) -> None:
        self.client = client
        self.collection = collection
        self.incremental = incremental
        self.chunk_size = chunk_size

        if incremental:
            self.cursor_field = incremental.cursor_path
            self.last_value = incremental.last_value
        else:
            self.cursor_column = None
            self.last_value = None

    @property
    def _sort_op(self) -> List[Optional[Tuple[str, int]]]:
        if not self.incremental or not self.last_value:
            return []

        if (
            self.incremental.row_order == "asc"
            and self.incremental.last_value_func is max
        ) or (
            self.incremental.row_order == "desc"
            and self.incremental.last_value_func is min
        ):
            return [(self.cursor_field, ASCENDING)]

        elif (
            self.incremental.row_order == "asc"
            and self.incremental.last_value_func is min
        ) or (
            self.incremental.row_order == "desc"
            and self.incremental.last_value_func is max
        ):
            return [(self.cursor_field, DESCENDING)]

        return []

    @property
    def _filter_op(self) -> Dict[str, Any]:
        """Build a filtering operator.

        Includes a field and the filtering condition for it.

        Returns:
            Dict[str, Any]: A dictionary with the filter operator.
        """
        if not (self.incremental and self.last_value):
            return {}

        filt = {}
        if self.incremental.last_value_func is max:
            filt = {self.cursor_field: {"$gte": self.last_value}}
            if self.incremental.end_value:
                filt[self.cursor_field]["$lt"] = self.incremental.end_value

        elif self.incremental.last_value_func is min:
            filt = {self.cursor_field: {"$lte": self.last_value}}
            if self.incremental.end_value:
                filt[self.cursor_field]["$gt"] = self.incremental.end_value

        return filt

    def _projection_op(
        self, projection: Optional[Union[Mapping[str, Any], Iterable[str]]]
    ) -> Optional[Dict[str, Any]]:
        """Build a projection operator.

        Args:
            projection (Optional[Union[Mapping[str, Any], Iterable[str]]]): A tuple of fields to include or a dict specifying fields to include or exclude.
            The incremental `primary_key` needs to be handle differently for inclusion
            and exclusion projections.

        Returns:
            Tuple[str, ...] | Dict[str, Any]: A tuple or dictionary with the projection operator.
        """
        if projection is None:
            return None

        projection_dict = dict(_fields_list_to_dict(projection, "projection"))

        if self.incremental:
            # this is an inclusion projection
            if any(v == 1 for v in projection_dict.values()):
                # ensure primary_key is included
                projection_dict.update(m={self.incremental.primary_key: 1})
            # this is an exclusion projection
            else:
                try:
                    # ensure primary_key isn't excluded
                    projection_dict.pop(self.incremental.primary_key)  # type: ignore
                except KeyError:
                    pass  # primary_key was properly not included in exclusion projection
                else:
                    dlt.common.logger.warn(
                        f"Primary key `{self.incremental.primary_key}` was removed from exclusion projection"
                    )

        return projection_dict

    def _limit(self, cursor: Cursor, limit: Optional[int] = None) -> TCursor:  # type: ignore
        """Apply a limit to the cursor, if needed.

        Args:
            cursor (Cursor): The cursor to apply the limit.
            limit (Optional[int]): The number of documents to load.

        Returns:
            Cursor: The cursor with the limit applied (if given).
        """
        if limit not in (0, None):
            if self.incremental is None or self.incremental.last_value_func is None:
                logger.warning(
                    "Using limit without ordering - results may be inconsistent."
                )

            cursor = cursor.limit(abs(limit))

        return cursor

    def load_documents(
        self,
        filter_: Dict[str, Any],
        limit: Optional[int] = None,
        projection: Optional[Union[Mapping[str, Any], Iterable[str]]] = None,
    ) -> Iterator[TDataItem]:
        """Construct the query and load the documents from the collection.

        Args:
            filter_ (Dict[str, Any]): The filter to apply to the collection.
            limit (Optional[int]): The number of documents to load.
            projection (Optional[Union[Mapping[str, Any], Iterable[str]]]): The projection to select fields to create the Cursor.

        Yields:
            Iterator[TDataItem]: An iterator of the loaded documents.
        """
        filter_op = self._filter_op
        _raise_if_intersection(filter_op, filter_)
        filter_op.update(filter_)

        projection_op = self._projection_op(projection)

        cursor = self.collection.find(filter=filter_op, projection=projection_op)
        if self._sort_op:
            cursor = cursor.sort(self._sort_op)

        cursor = self._limit(cursor, limit)

        while docs_slice := list(islice(cursor, self.chunk_size)):
            res = map_nested_in_place(convert_mongo_objs, docs_slice)
            if len(res) > 0 and "_id" in res[0] and isinstance(res[0]["_id"], dict):
                yield dlt.mark.with_hints(
                    res,
                    dlt.mark.make_hints(columns={"_id": {"data_type": "json"}}),
                )
            else:
                yield res


class CollectionLoaderParallel(CollectionLoader):
    def _get_document_count(self) -> int:
        return self.collection.count_documents(filter=self._filter_op)

    def _create_batches(self, limit: Optional[int] = None) -> List[Dict[str, int]]:
        doc_count = self._get_document_count()
        if limit:
            doc_count = min(doc_count, abs(limit))

        batches = []
        left_to_load = doc_count

        for sk in range(0, doc_count, self.chunk_size):
            batches.append(dict(skip=sk, limit=min(self.chunk_size, left_to_load)))
            left_to_load -= self.chunk_size

        return batches

    def _get_cursor(
        self,
        filter_: Dict[str, Any],
        projection: Optional[Union[Mapping[str, Any], Iterable[str]]] = None,
    ) -> TCursor:
        """Get a reading cursor for the collection.

        Args:
            filter_ (Dict[str, Any]): The filter to apply to the collection.
            projection (Optional[Union[Mapping[str, Any], Iterable[str]]]): The projection to select fields to create the Cursor.

        Returns:
            Cursor: The cursor for the collection.
        """
        filter_op = self._filter_op
        _raise_if_intersection(filter_op, filter_)
        filter_op.update(filter_)

        projection_op = self._projection_op(projection)

        cursor = self.collection.find(filter=filter_op, projection=projection_op)
        if self._sort_op:
            cursor = cursor.sort(self._sort_op)

        return cursor

    @dlt.defer
    def _run_batch(self, cursor: TCursor, batch: Dict[str, int]) -> TDataItem:
        cursor = cursor.clone()

        data = []
        for document in cursor.skip(batch["skip"]).limit(batch["limit"]):
            data.append(map_nested_in_place(convert_mongo_objs, document))

        return data

    def _get_all_batches(
        self,
        filter_: Dict[str, Any],
        limit: Optional[int] = None,
        projection: Optional[Union[Mapping[str, Any], Iterable[str]]] = None,
    ) -> Iterator[TDataItem]:
        """Load all documents from the collection in parallel batches.

        Args:
            filter_ (Dict[str, Any]): The filter to apply to the collection.
            limit (Optional[int]): The maximum number of documents to load.
            projection (Optional[Union[Mapping[str, Any], Iterable[str]]]): The projection to select fields to create the Cursor.

        Yields:
            Iterator[TDataItem]: An iterator of the loaded documents.
        """
        batches = self._create_batches(limit=limit)
        cursor = self._get_cursor(filter_=filter_, projection=projection)

        for batch in batches:
            yield self._run_batch(cursor=cursor, batch=batch)

    def load_documents(
        self,
        filter_: Dict[str, Any],
        limit: Optional[int] = None,
        projection: Optional[Union[Mapping[str, Any], Iterable[str]]] = None,
    ) -> Iterator[TDataItem]:
        """Load documents from the collection in parallel.

        Args:
            filter_ (Dict[str, Any]): The filter to apply to the collection.
            limit (Optional[int]): The number of documents to load.
            projection (Optional[Union[Mapping[str, Any], Iterable[str]]]): The projection to select fields to create the Cursor.

        Yields:
            Iterator[TDataItem]: An iterator of the loaded documents.
        """
        for document in self._get_all_batches(
            limit=limit, filter_=filter_, projection=projection
        ):
            yield document


class CollectionArrowLoader(CollectionLoader):
    """
    Mongo DB collection loader, which uses
    Apache Arrow for data processing.
    """

    def load_documents(
        self,
        filter_: Dict[str, Any],
        limit: Optional[int] = None,
        projection: Optional[Union[Mapping[str, Any], Iterable[str]]] = None,
        pymongoarrow_schema: Any = None,
    ) -> Iterator[Any]:
        """
        Load documents from the collection in Apache Arrow format.

        Args:
            filter_ (Dict[str, Any]): The filter to apply to the collection.
            limit (Optional[int]): The number of documents to load.
            projection (Optional[Union[Mapping[str, Any], Iterable[str]]]): The projection to select fields to create the Cursor.
            pymongoarrow_schema (Any): The mapping of field types to convert BSON to Arrow.

        Yields:
            Iterator[Any]: An iterator of the loaded documents.
        """
        from pymongoarrow.context import PyMongoArrowContext  # type: ignore
        from pymongoarrow.lib import process_bson_stream  # type: ignore

        filter_op = self._filter_op
        _raise_if_intersection(filter_op, filter_)
        filter_op.update(filter_)

        projection_op = self._projection_op(projection)

        # NOTE the `filter_op` isn't passed
        cursor = self.collection.find_raw_batches(
            filter_, batch_size=self.chunk_size, projection=projection_op
        )
        if self._sort_op:
            cursor = cursor.sort(self._sort_op)  # type: ignore

        cursor = self._limit(cursor, limit)  # type: ignore

        context = PyMongoArrowContext.from_schema(
            schema=pymongoarrow_schema, codec_options=self.collection.codec_options
        )
        for batch in cursor:
            process_bson_stream(batch, context)
            table = context.finish()
            yield convert_arrow_columns(table)


class CollectionArrowLoaderParallel(CollectionLoaderParallel):
    """
    Mongo DB collection parallel loader, which uses
    Apache Arrow for data processing.
    """

    def load_documents(
        self,
        filter_: Dict[str, Any],
        limit: Optional[int] = None,
        projection: Optional[Union[Mapping[str, Any], Iterable[str]]] = None,
        pymongoarrow_schema: Any = None,
    ) -> Iterator[TDataItem]:
        """Load documents from the collection in parallel.

        Args:
            filter_ (Dict[str, Any]): The filter to apply to the collection.
            limit (Optional[int]): The number of documents to load.
            projection (Optional[Union[Mapping[str, Any], Iterable[str]]]): The projection to select fields to create the Cursor.
            pymongoarrow_schema (Any): The mapping of field types to convert BSON to Arrow.

        Yields:
            Iterator[TDataItem]: An iterator of the loaded documents.
        """
        yield from self._get_all_batches(
            limit=limit,
            filter_=filter_,
            projection=projection,
            pymongoarrow_schema=pymongoarrow_schema,
        )

    def _get_all_batches(
        self,
        filter_: Dict[str, Any],
        limit: Optional[int] = None,
        projection: Optional[Union[Mapping[str, Any], Iterable[str]]] = None,
        pymongoarrow_schema: Any = None,
    ) -> Iterator[TDataItem]:
        """Load all documents from the collection in parallel batches.

        Args:
            filter_ (Dict[str, Any]): The filter to apply to the collection.
            limit (Optional[int]): The maximum number of documents to load.
            projection (Optional[Union[Mapping[str, Any], Iterable[str]]]): The projection to select fields to create the Cursor.
            pymongoarrow_schema (Any): The mapping of field types to convert BSON to Arrow.

        Yields:
            Iterator[TDataItem]: An iterator of the loaded documents.
        """
        batches = self._create_batches(limit=limit)
        cursor = self._get_cursor(filter_=filter_, projection=projection)
        for batch in batches:
            yield self._run_batch(
                cursor=cursor,
                batch=batch,
                pymongoarrow_schema=pymongoarrow_schema,
            )

    def _get_cursor(
        self,
        filter_: Dict[str, Any],
        projection: Optional[Union[Mapping[str, Any], Iterable[str]]] = None,
    ) -> TCursor:
        """Get a reading cursor for the collection.

        Args:
            filter_ (Dict[str, Any]): The filter to apply to the collection.
            projection (Optional[Union[Mapping[str, Any], Iterable[str]]]): The projection to select fields to create the Cursor.

        Returns:
            Cursor: The cursor for the collection.
        """
        filter_op = self._filter_op
        _raise_if_intersection(filter_op, filter_)
        filter_op.update(filter_)

        projection_op = self._projection_op(projection)

        cursor = self.collection.find_raw_batches(
            filter=filter_op, batch_size=self.chunk_size, projection=projection_op
        )
        if self._sort_op:
            cursor = cursor.sort(self._sort_op)  # type: ignore

        return cursor

    @dlt.defer
    def _run_batch(
        self,
        cursor: TCursor,
        batch: Dict[str, int],
        pymongoarrow_schema: Any = None,
    ) -> TDataItem:
        from pymongoarrow.context import PyMongoArrowContext
        from pymongoarrow.lib import process_bson_stream

        cursor = cursor.clone()

        context = PyMongoArrowContext.from_schema(
            schema=pymongoarrow_schema, codec_options=self.collection.codec_options
        )
        for chunk in cursor.skip(batch["skip"]).limit(batch["limit"]):
            process_bson_stream(chunk, context)
            table = context.finish()
            yield convert_arrow_columns(table)


class CollectionAggregationLoader(CollectionLoader):
    """
    MongoDB collection loader that uses aggregation pipelines instead of find queries.
    """

    def __init__(
        self,
        client: TMongoClient,
        collection: TCollection,
        chunk_size: int,
        incremental: Optional[dlt.sources.incremental[Any]] = None,
    ) -> None:
        super().__init__(client, collection, chunk_size, incremental)
        self.custom_query: Optional[List[Dict[str, Any]]] = None

    def set_custom_query(self, query: List[Dict[str, Any]]):
        """Set the custom aggregation pipeline query"""
        self.custom_query = query

    def load_documents(
        self,
        filter_: Dict[str, Any],
        limit: Optional[int] = None,
        projection: Optional[Union[Mapping[str, Any], Iterable[str]]] = None,
    ) -> Iterator[TDataItem]:
        """Load documents using aggregation pipeline"""
        if not self.custom_query:
            # Fallback to parent method if no custom query
            yield from super().load_documents(filter_, limit, projection)
            return

        # Build aggregation pipeline
        pipeline = list(self.custom_query)  # Copy the query

        # For custom queries, we assume incremental filtering is already handled
        # via interval placeholders (:interval_start, :interval_end) in the query itself.
        # We don't add additional incremental filtering to avoid conflicts.

        # Add additional filter if provided
        if filter_:
            filter_match = {"$match": filter_}
            pipeline.insert(0, filter_match)

        # Add limit if specified
        if limit and limit > 0:
            pipeline.append({"$limit": limit})

        # Add maxTimeMS to prevent hanging
        cursor = self.collection.aggregate(
            pipeline,
            allowDiskUse=True,
            batchSize=min(self.chunk_size, 101),
            maxTimeMS=30000,  # 30 second timeout
        )

        docs_buffer = []
        try:
            for doc in cursor:
                docs_buffer.append(doc)

                if len(docs_buffer) >= self.chunk_size:
                    res = map_nested_in_place(convert_mongo_objs, docs_buffer)
                    if (
                        len(res) > 0
                        and "_id" in res[0]
                        and isinstance(res[0]["_id"], dict)
                    ):
                        yield dlt.mark.with_hints(
                            res,
                            dlt.mark.make_hints(columns={"_id": {"data_type": "json"}}),
                        )
                    else:
                        yield res
                    docs_buffer = []

            # Yield any remaining documents
            if docs_buffer:
                res = map_nested_in_place(convert_mongo_objs, docs_buffer)
                if len(res) > 0 and "_id" in res[0] and isinstance(res[0]["_id"], dict):
                    yield dlt.mark.with_hints(
                        res,
                        dlt.mark.make_hints(columns={"_id": {"data_type": "json"}}),
                    )
                else:
                    yield res
        finally:
            cursor.close()


class CollectionAggregationLoaderParallel(CollectionAggregationLoader):
    """
    MongoDB collection parallel loader that uses aggregation pipelines.
    Note: Parallel loading is not supported for aggregation pipelines due to cursor limitations.
    Falls back to sequential loading.
    """

    def load_documents(
        self,
        filter_: Dict[str, Any],
        limit: Optional[int] = None,
        projection: Optional[Union[Mapping[str, Any], Iterable[str]]] = None,
    ) -> Iterator[TDataItem]:
        """Load documents using aggregation pipeline (sequential only)"""
        logger.warning(
            "Parallel loading is not supported for MongoDB aggregation pipelines. Using sequential loading."
        )
        yield from super().load_documents(filter_, limit, projection)


class CollectionAggregationArrowLoader(CollectionAggregationLoader):
    """
    MongoDB collection aggregation loader that uses Apache Arrow for data processing.
    """

    def load_documents(
        self,
        filter_: Dict[str, Any],
        limit: Optional[int] = None,
        projection: Optional[Union[Mapping[str, Any], Iterable[str]]] = None,
        pymongoarrow_schema: Any = None,
    ) -> Iterator[Any]:
        """Load documents using aggregation pipeline with Arrow format"""
        logger.warning(
            "Arrow format is not directly supported for MongoDB aggregation pipelines. Converting to Arrow after loading."
        )

        # Load documents normally and convert to arrow format
        for batch in super().load_documents(filter_, limit, projection):
            if batch:  # Only process non-empty batches
                try:
                    from dlt.common.libs.pyarrow import pyarrow

                    # Convert dict batch to arrow table
                    table = pyarrow.Table.from_pylist(batch)
                    yield convert_arrow_columns(table)
                except ImportError:
                    logger.warning(
                        "PyArrow not available, falling back to object format"
                    )
                    yield batch


class CollectionAggregationArrowLoaderParallel(CollectionAggregationArrowLoader):
    """
    MongoDB collection parallel aggregation loader with Arrow support.
    Falls back to sequential loading.
    """

    def load_documents(
        self,
        filter_: Dict[str, Any],
        limit: Optional[int] = None,
        projection: Optional[Union[Mapping[str, Any], Iterable[str]]] = None,
        pymongoarrow_schema: Any = None,
    ) -> Iterator[TDataItem]:
        """Load documents using aggregation pipeline with Arrow format (sequential only)"""
        logger.warning(
            "Parallel loading is not supported for MongoDB aggregation pipelines. Using sequential loading."
        )
        yield from super().load_documents(
            filter_, limit, projection, pymongoarrow_schema
        )


def collection_documents(
    client: TMongoClient,
    collection: TCollection,
    filter_: Dict[str, Any],
    projection: Union[Dict[str, Any], List[str]],
    pymongoarrow_schema: "pymongoarrow.schema.Schema",
    incremental: Optional[dlt.sources.incremental[Any]] = None,
    parallel: bool = False,
    limit: Optional[int] = None,
    chunk_size: Optional[int] = 10000,
    data_item_format: Optional[TDataItemFormat] = "object",
    custom_query: Optional[List[Dict[str, Any]]] = None,
) -> Iterator[TDataItem]:
    """
    A DLT source which loads data from a Mongo database using PyMongo.
    Resources are automatically created for the collection.

    Args:
        client (MongoClient): The PyMongo client `pymongo.MongoClient` instance.
        collection (Collection): The collection `pymongo.collection.Collection` to load.
        filter_ (Dict[str, Any]): The filter to apply to the collection.
        projection (Optional[Union[Mapping[str, Any], Iterable[str]]]): The projection to select fields to create the Cursor.
            when loading the collection. Supported inputs:
                include (list) - ["year", "title"]
                include (dict) - {"year": True, "title": True}
                exclude (dict) - {"released": False, "runtime": False}
            Note: Can't mix include and exclude statements '{"title": True, "released": False}`
        pymongoarrow_schema (pymongoarrow.schema.Schema): The mapping of field types to convert BSON to Arrow.
        incremental (Optional[dlt.sources.incremental[Any]]): The incremental configuration.
        parallel (bool): Option to enable parallel loading for the collection. Default is False.
        limit (Optional[int]): The maximum number of documents to load.
        chunk_size (Optional[int]): The number of documents to load in each batch.
        data_item_format (Optional[TDataItemFormat]): The data format to use for loading.
            Supported formats:
                object - Python objects (dicts, lists).
                arrow - Apache Arrow tables.
        custom_query (Optional[List[Dict[str, Any]]]): Custom MongoDB aggregation pipeline to execute instead of find()

    Returns:
        Iterable[DltResource]: A list of DLT resources for each collection to be loaded.
    """
    if data_item_format == "arrow" and not PYMONGOARROW_AVAILABLE:
        dlt.common.logger.warn(
            "'pymongoarrow' is not installed; falling back to standard MongoDB CollectionLoader."
        )
        data_item_format = "object"

    if data_item_format != "arrow" and pymongoarrow_schema:
        dlt.common.logger.warn(
            "Received value for `pymongoarrow_schema`, but `data_item_format=='object'` "
            "Use `data_item_format=='arrow'` to enforce schema."
        )

    if data_item_format == "arrow" and pymongoarrow_schema and projection:
        dlt.common.logger.warn(
            "Received values for both `pymongoarrow_schema` and `projection`. Since both "
            "create a projection to select fields, `projection` will be ignored."
        )

    # If custom query is provided, use aggregation loaders
    if custom_query:
        if parallel:
            if data_item_format == "arrow":
                LoaderClass = CollectionAggregationArrowLoaderParallel
            else:
                LoaderClass = CollectionAggregationLoaderParallel  # type: ignore
        else:
            if data_item_format == "arrow":
                LoaderClass = CollectionAggregationArrowLoader  # type: ignore
            else:
                LoaderClass = CollectionAggregationLoader  # type: ignore
    else:
        if parallel:
            if data_item_format == "arrow":
                LoaderClass = CollectionArrowLoaderParallel
            else:
                LoaderClass = CollectionLoaderParallel  # type: ignore
        else:
            if data_item_format == "arrow":
                LoaderClass = CollectionArrowLoader  # type: ignore
            else:
                LoaderClass = CollectionLoader  # type: ignore

    loader = LoaderClass(
        client, collection, incremental=incremental, chunk_size=chunk_size
    )

    # Set custom query if provided
    if custom_query and hasattr(loader, "set_custom_query"):
        loader.set_custom_query(custom_query)

    # Load documents based on loader type
    if isinstance(
        loader,
        (
            CollectionArrowLoader,
            CollectionArrowLoaderParallel,
            CollectionAggregationArrowLoader,
            CollectionAggregationArrowLoaderParallel,
        ),
    ):
        yield from loader.load_documents(
            limit=limit,
            filter_=filter_,
            projection=projection,
            pymongoarrow_schema=pymongoarrow_schema,
        )
    else:
        yield from loader.load_documents(
            limit=limit, filter_=filter_, projection=projection
        )


def convert_mongo_objs(value: Any) -> Any:
    """MongoDB to dlt type conversion when using Python loaders.

    Notes:
        The method `ObjectId.__str__()` creates a hexstring using `binascii.hexlify(__id).decode()`

    """
    if isinstance(value, (ObjectId, Decimal128)):
        return str(value)
    if isinstance(value, _datetime.datetime):
        return ensure_pendulum_datetime(value)
    if isinstance(value, Regex):
        return value.try_compile().pattern
    if isinstance(value, Timestamp):
        date = value.as_datetime()
        return ensure_pendulum_datetime(date)

    return value


def convert_arrow_columns(table: Any) -> Any:
    """Convert the given table columns to Python types.

    Notes:
        Calling str() matches the `convert_mongo_obs()` used in non-arrow code.
        Pymongoarrow converts ObjectId to `fixed_size_binary[12]`, which can't be
        converted to a string as a vectorized operation because it contains ASCII characters.

        Instead, you need to loop over values using:
        ```python
        pyarrow.array([v.as_buffer().hex() for v in object_id_array], type=pyarrow.string())
        # pymongoarrow simplifies this by allowing this syntax
        [str(v) for v in object_id_array]
        ```

    Args:
        table (pyarrow.lib.Table): The table to convert.

    Returns:
        pyarrow.lib.Table: The table with the columns converted.
    """
    from dlt.common.libs.pyarrow import pyarrow
    from pymongoarrow.types import (  # type: ignore
        _is_binary,
        _is_code,
        _is_decimal128,
        _is_objectid,
    )

    for i, field in enumerate(table.schema):
        if _is_objectid(field.type) or _is_decimal128(field.type):
            col_values = [str(value) for value in table[field.name]]
            table = table.set_column(
                i,
                pyarrow.field(field.name, pyarrow.string()),
                pyarrow.array(col_values, type=pyarrow.string()),
            )
        else:
            type_ = None
            if _is_binary(field.type):
                type_ = pyarrow.binary()
            elif _is_code(field.type):
                type_ = pyarrow.string()

            if type_:
                col_values = [value.as_py() for value in table[field.name]]
                table = table.set_column(
                    i,
                    pyarrow.field(field.name, type_),
                    pyarrow.array(col_values, type=type_),
                )
    return table


def client_from_credentials(connection_url: str) -> TMongoClient:
    client: TMongoClient = MongoClient(
        connection_url, uuidRepresentation="standard", tz_aware=True
    )
    return client


def _raise_if_intersection(filter1: Dict[str, Any], filter2: Dict[str, Any]) -> None:
    """
    Raise an exception, if the given filters'
    fields are intersecting.

    Args:
        filter1 (Dict[str, Any]): The first filter.
        filter2 (Dict[str, Any]): The second filter.
    """
    field_inter = filter1.keys() & filter2.keys()
    for field in field_inter:
        if filter1[field].keys() & filter2[field].keys():
            str_repr = str({field: filter1[field]})
            raise ValueError(
                (
                    f"Filtering operator {str_repr} is already used by the "
                    "incremental and can't be used in the filter."
                )
            )


@configspec
class MongoDbCollectionConfiguration(BaseConfiguration):
    incremental: Optional[dlt.sources.incremental] = None  # type: ignore[type-arg]


@configspec
class MongoDbCollectionResourceConfiguration(BaseConfiguration):
    connection_url: dlt.TSecretValue = dlt.secrets.value
    database: Optional[str] = dlt.config.value
    collection: str = dlt.config.value
    incremental: Optional[dlt.sources.incremental] = None  # type: ignore[type-arg]
    write_disposition: Optional[str] = dlt.config.value
    parallel: Optional[bool] = False
    projection: Optional[Union[Mapping[str, Any], Iterable[str]]] = dlt.config.value


def convert_mongo_shell_to_extended_json(query_string: str) -> str:
    """
    Convert MongoDB shell syntax to MongoDB Extended JSON v2 format.

    This function handles common MongoDB shell constructs like ISODate, ObjectId,
    NumberLong, NumberDecimal, etc. and converts them to their Extended JSON equivalents
    that can be parsed by bson.json_util.

    Args:
        query_string: A string containing MongoDB shell syntax

    Returns:
        A string with MongoDB Extended JSON v2 format

    Examples:
        >>> convert_mongo_shell_to_extended_json('ISODate("2010-01-01T00:00:00.000Z")')
        '{"$date": "2010-01-01T00:00:00.000Z"}'

        >>> convert_mongo_shell_to_extended_json('ObjectId("507f1f77bcf86cd799439011")')
        '{"$oid": "507f1f77bcf86cd799439011"}'
    """
    converted = query_string

    # Convert ISODate("...") to {"$date": "..."}
    # Pattern matches ISODate("2010-01-01T00:00:00.000+0000") or similar
    converted = re.sub(r'ISODate\("([^"]+)"\)', r'{"$date": "\1"}', converted)

    # Convert ObjectId("...") to {"$oid": "..."}
    converted = re.sub(r'ObjectId\("([^"]+)"\)', r'{"$oid": "\1"}', converted)

    # Convert NumberLong(...) to {"$numberLong": "..."}
    # Note: NumberLong can have quotes or not: NumberLong(123) or NumberLong("123")
    converted = re.sub(r'NumberLong\("([^"]+)"\)', r'{"$numberLong": "\1"}', converted)
    converted = re.sub(r"NumberLong\(([^)]+)\)", r'{"$numberLong": "\1"}', converted)

    # Convert NumberInt(...) to {"$numberInt": "..."}
    converted = re.sub(r'NumberInt\("([^"]+)"\)', r'{"$numberInt": "\1"}', converted)
    converted = re.sub(r"NumberInt\(([^)]+)\)", r'{"$numberInt": "\1"}', converted)

    # Convert NumberDecimal("...") to {"$numberDecimal": "..."}
    converted = re.sub(
        r'NumberDecimal\("([^"]+)"\)', r'{"$numberDecimal": "\1"}', converted
    )

    # Convert Timestamp(..., ...) to {"$timestamp": {"t": ..., "i": ...}}
    # Timestamp(1234567890, 1) -> {"$timestamp": {"t": 1234567890, "i": 1}}
    converted = re.sub(
        r"Timestamp\((\d+),\s*(\d+)\)", r'{"$timestamp": {"t": \1, "i": \2}}', converted
    )

    # Convert BinData(..., "...") to {"$binary": {"base64": "...", "subType": "..."}}
    converted = re.sub(
        r'BinData\((\d+),\s*"([^"]+)"\)',
        r'{"$binary": {"base64": "\2", "subType": "\1"}}',
        converted,
    )

    # Convert MinKey() to {"$minKey": 1}
    converted = re.sub(r"MinKey\(\)", r'{"$minKey": 1}', converted)

    # Convert MaxKey() to {"$maxKey": 1}
    converted = re.sub(r"MaxKey\(\)", r'{"$maxKey": 1}', converted)

    # Convert UUID("...") to {"$uuid": "..."}
    converted = re.sub(r'UUID\("([^"]+)"\)', r'{"$uuid": "\1"}', converted)

    # Convert DBRef("collection", "id") to {"$ref": "collection", "$id": "id"}
    converted = re.sub(
        r'DBRef\("([^"]+)",\s*"([^"]+)"\)', r'{"$ref": "\1", "$id": "\2"}', converted
    )

    # Convert Code("...") to {"$code": "..."}
    converted = re.sub(r'Code\("([^"]+)"\)', r'{"$code": "\1"}', converted)

    return converted


__source_name__ = "mongodb"


# MongoDB destination helper functions
def process_file_items(file_path: str) -> list[dict]:
    """Process items from a file path (JSONL format)."""
    import json

    documents = []
    with open(file_path, "r") as f:
        for line in f:
            if line.strip():
                doc = json.loads(line.strip())
                documents.append(doc)  # Include all fields including DLT metadata
    return documents



"""Primer API client helpers"""

from typing import Any, Dict, List, Optional

from dlt.common.pendulum import pendulum
from dlt.sources.helpers import requests


def build_date_params(
    start_date: Optional[pendulum.DateTime] = None,
    end_date: Optional[pendulum.DateTime] = None,
) -> Dict[str, str]:
    params: Dict[str, str] = {}
    if start_date:
        params["from_date"] = start_date.start_of("day").to_iso8601_string()
    if end_date:
        params["to_date"] = end_date.add(days=1).start_of("day").to_iso8601_string()
    return params


class PrimerApi:
    """
    A Primer API client that handles pagination and data fetching.
    """

    BASE_URL = "https://api.primer.io"

    def __init__(
        self,
        api_key: str,
        api_version: str = "2.4",
    ) -> None:
        self.api_key = api_key
        self.api_version = api_version

    def _get_headers(self) -> Dict[str, str]:
        return {
            "X-API-KEY": self.api_key,
            "X-API-VERSION": self.api_version,
        }

    def list_payment_ids(
        self,
        start_date: Optional[pendulum.DateTime] = None,
        end_date: Optional[pendulum.DateTime] = None,
    ) -> List[str]:
        """List all payment IDs from the payments endpoint."""
        url = f"{self.BASE_URL}/payments"
        params: Dict[str, Any] = build_date_params(start_date, end_date)

        payment_ids: List[str] = []

        while True:
            response = requests.get(url, params=params, headers=self._get_headers())
            response.raise_for_status()

            json_data = response.json()
            data = json_data.get("data", [])

            if len(data) == 0:
                break

            for item in data:
                payment_ids.append(item["id"])

            next_cursor = json_data.get("nextCursor")
            if not next_cursor:
                break

            params["cursor"] = next_cursor

        return payment_ids

    def get_payment(self, payment_id: str) -> Dict[str, Any]:
        """Get payment by ID."""
        url = f"{self.BASE_URL}/payments/{payment_id}"
        response = requests.get(url, headers=self._get_headers())
        response.raise_for_status()
        return response.json()



# Copyright 2022-2025 ScaleVector
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A source loading player profiles and games from chess.com api"""

from typing import Any, Callable, Dict, Iterator, List, Sequence

import dlt
from dlt.common import pendulum
from dlt.common.typing import TDataItem
from dlt.sources import DltResource
from dlt.sources.helpers import requests

from .helpers import get_path_with_retry, get_url_with_retry, validate_month_string
from .settings import UNOFFICIAL_CHESS_API_URL


@dlt.source(name="chess", max_table_nesting=0)
def source(
    players: List[str], start_month: str = None, end_month: str = None
) -> Sequence[DltResource]:
    """
    A dlt source for the chess.com api. It groups several resources (in this case chess.com API endpoints) containing
    various types of data: user profiles or chess match results
    Args:
        players (List[str]): A list of the player usernames for which to get the data.
        start_month (str, optional): Filters out all the matches happening before `start_month`. Defaults to None.
        end_month (str, optional): Filters out all the matches happening after `end_month`. Defaults to None.
    Returns:
        Sequence[DltResource]: A sequence of resources that can be selected from including players_profiles,
        players_archives, players_games, players_online_status
    """
    return (
        players_profiles(players),
        players_archives(players),
        players_games(players, start_month=start_month, end_month=end_month),
        players_online_status(players),
    )


@dlt.resource(
    write_disposition="replace",
    columns={
        "last_online": {"data_type": "timestamp"},
        "joined": {"data_type": "timestamp"},
    },
)
def players_profiles(players: List[str]) -> Iterator[TDataItem]:
    """
    Yields player profiles for a list of player usernames.
    Args:
        players (List[str]): List of player usernames to retrieve profiles for.
    Yields:
        Iterator[TDataItem]: An iterator over player profiles data.
    """

    # get archives in parallel by decorating the http request with defer
    @dlt.defer
    def _get_profile(username: str) -> TDataItem:
        return get_path_with_retry(f"player/{username}")

    for username in players:
        yield _get_profile(username)


@dlt.resource(write_disposition="replace", selected=False)
def players_archives(players: List[str]) -> Iterator[List[TDataItem]]:
    """
    Yields url to game archives for specified players.
    Args:
        players (List[str]): List of player usernames to retrieve archives for.
    Yields:
        Iterator[List[TDataItem]]: An iterator over list of player archive data.
    """
    for username in players:
        data = get_path_with_retry(f"player/{username}/games/archives")
        yield data.get("archives", [])


@dlt.resource(
    write_disposition="replace", columns={"end_time": {"data_type": "timestamp"}}
)
def players_games(
    players: List[str], start_month: str = None, end_month: str = None
) -> Iterator[Callable[[], List[TDataItem]]]:
    """
    Yields `players` games that happened between `start_month` and `end_month`.
    Args:
        players (List[str]): List of player usernames to retrieve games for.
        start_month (str, optional): The starting month in the format "YYYY/MM". Defaults to None.
        end_month (str, optional): The ending month in the format "YYYY/MM". Defaults to None.
    Yields:
        Iterator[Callable[[], List[TDataItem]]]: An iterator over callables that return a list of games for each player.
    """  # do a simple validation to prevent common mistakes in month format
    validate_month_string(start_month)
    validate_month_string(end_month)

    # get a list of already checked archives
    # from your point of view, the state is python dictionary that will have the same content the next time this function is called
    checked_archives = dlt.current.resource_state().setdefault("archives", [])
    # get player archives, note that you can call the resource like any other function and just iterate it like a list
    archives = players_archives(players)

    # get archives in parallel by decorating the http request with defer
    @dlt.defer
    def _get_archive(url: str) -> List[TDataItem]:
        try:
            games = get_url_with_retry(url).get("games", [])
            return games  # type: ignore
        except requests.HTTPError as http_err:
            # sometimes archives are not available and the error seems to be permanent
            if http_err.response.status_code == 404:
                return []
            raise

    # enumerate the archives
    for url in archives:
        # the `url` format is https://api.chess.com/pub/player/{username}/games/{YYYY}/{MM}
        if start_month and url[-7:] < start_month:
            continue
        if end_month and url[-7:] > end_month:
            continue
        # do not download archive again
        if url in checked_archives:
            continue
        checked_archives.append(url)
        # get the filtered archive
        yield _get_archive(url)


@dlt.resource(write_disposition="append")
def players_online_status(players: List[str]) -> Iterator[TDataItem]:
    """
    Returns current online status for a list of players.
    Args:
        players (List[str]): List of player usernames to check online status for.
    Yields:
        Iterator[TDataItem]: An iterator over the online status of each player.
    """
    # we'll use unofficial endpoint to get online status, the official seems to be removed
    for player in players:
        status = get_url_with_retry(f"{UNOFFICIAL_CHESS_API_URL}user/popup/{player}")
        # return just relevant selection
        yield {
            "username": player,
            "onlineStatus": status["onlineStatus"],
            "lastLoginDate": status["lastLoginDate"],
            "check_time": pendulum.now(),  # dlt can deal with native python dates
        }


@dlt.source
def chess_dlt_config_example(
    secret_str: str = dlt.secrets.value,
    secret_dict: Dict[str, Any] = dlt.secrets.value,
    config_int: int = dlt.config.value,
) -> DltResource:
    """
    An example of a source that uses dlt to provide secrets and config values.
    Args:
        secret_str (str, optional): Secret string provided by dlt.secrets.value. Defaults to dlt.secrets.value.
        secret_dict (Dict[str, Any], optional): Secret dictionary provided by dlt.secrets.value. Defaults to dlt.secrets.value.
        config_int (int, optional): Config integer provided by dlt.config.value. Defaults to dlt.config.value.
    Returns:
        DltResource: Returns a resource yielding the configured values.
    """

    # returns a resource yielding the configured values - it is just a test
    return dlt.resource([secret_str, secret_dict, config_int], name="config_values")



from typing import Iterable, Optional

from ..http_client import create_client


class ClickupClient:
    def __init__(self, api_token: str):
        self.session = create_client()
        self.base_url = "https://api.clickup.com/api/v2"
        self.headers = {"Authorization": api_token}

    def get(self, endpoint: str, params: Optional[dict] = None) -> dict:
        url = f"{self.base_url}{endpoint}"
        resp = self.session.get(url, headers=self.headers, params=params or {})
        resp.raise_for_status()
        return resp.json()

    def paginated(
        self, endpoint: str, key: str, params: Optional[dict] = None
    ) -> Iterable[dict]:
        page = 0
        params = params or {}
        while True:
            params["page"] = page
            data = self.get(endpoint, params)
            items = data.get(key, data)
            if not items:
                break
            for item in items:
                yield item
            if data.get("last_page") or len(items) < params.get("page_size", 100):
                break
            page += 1

    def get_teams(self):
        data = self.get("/team")
        return data.get("teams", [])

    def get_spaces(self):
        for team in self.get_teams():
            for space in self.paginated(f"/team/{team['id']}/space", "spaces"):
                yield space

    def get_lists(self):
        for space in self.get_spaces():
            for lst in self.paginated(f"/space/{space['id']}/list", "lists"):
                yield lst



# Copyright 2022-2025 ScaleVector
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A source to extract Kafka messages.

When extraction starts, partitions length is checked -
data is read only up to it, overriding the default Kafka's
behavior of waiting for new messages in endless loop.
"""

from contextlib import closing
from typing import Any, Callable, Dict, Iterable, List, Optional, Union

import dlt
from confluent_kafka import Consumer, Message  # type: ignore
from dlt.common import logger
from dlt.common.time import ensure_pendulum_datetime
from dlt.common.typing import TAnyDateTime, TDataItem

from .helpers import (
    KafkaCredentials,
    OffsetTracker,
    default_msg_processor,
)


@dlt.resource(
    name="kafka_messages",
    table_name=lambda msg: msg["_kafka"]["topic"],
    standalone=True,
)
def kafka_consumer(
    topics: Union[str, List[str]],
    credentials: Union[KafkaCredentials, Consumer] = dlt.secrets.value,
    msg_processor: Optional[
        Callable[[Message], Dict[str, Any]]
    ] = default_msg_processor,
    batch_size: Optional[int] = 3000,
    batch_timeout: Optional[int] = 3,
    start_from: Optional[TAnyDateTime] = None,
) -> Iterable[TDataItem]:
    """Extract recent messages from the given Kafka topics.

    The resource tracks offsets for all the topics and partitions,
    and so reads data incrementally.

    Messages from different topics are saved in different tables.

    Args:
        topics (Union[str, List[str]]): Names of topics to extract.
        credentials (Optional[Union[KafkaCredentials, Consumer]]):
            Auth credentials or an initiated Kafka consumer. By default,
            is taken from secrets.
        msg_processor(Optional[Callable]): A function-converter,
            which'll process every Kafka message after it's read and
            before it's transfered to the destination.
        batch_size (Optional[int]): Messages batch size to read at once.
        batch_timeout (Optional[int]): Maximum time to wait for a batch
            consume, in seconds.
        start_from (Optional[TAnyDateTime]): A timestamp, at which to start
            reading. Older messages are ignored.

    Yields:
        Iterable[TDataItem]: Kafka messages.
    """
    if not isinstance(topics, list):
        topics = [topics]

    if isinstance(credentials, Consumer):
        consumer = credentials
    elif isinstance(credentials, KafkaCredentials):
        consumer = credentials.init_consumer()
    else:
        raise TypeError(
            (
                "Wrong credentials type provided. Need to be of type: "
                "KafkaCredentials or confluent_kafka.Consumer"
            )
        )

    if start_from is not None:
        start_from = ensure_pendulum_datetime(start_from)

    tracker = OffsetTracker(consumer, topics, dlt.current.resource_state(), start_from)  # type: ignore

    # read messages up to the maximum offsets,
    # not waiting for new messages
    with closing(consumer):
        while True:
            messages = consumer.consume(batch_size, timeout=batch_timeout)
            if not messages:
                break

            batch = []
            for msg in messages:
                if msg.error():
                    err = msg.error()
                    if err.retriable() or not err.fatal():
                        logger.warning(f"ERROR: {err} - RETRYING")
                    else:
                        raise err
                else:
                    batch.append(msg_processor(msg))  # type: ignore
                    tracker.renew(msg)

            yield batch

            if tracker.has_unread is False:
                return



"""Monday.com source settings and constants"""

# GraphQL query for fetching app installs
APP_INSTALLS_QUERY = """
query ($app_id: ID!, $account_id: ID, $limit: Int!, $page: Int!) {
    app_installs(
        app_id: $app_id
        account_id: $account_id
        limit: $limit
        page: $page
    ) {
        app_id
        timestamp
        app_install_account {
            id
        }
        app_install_user {
            id
        }
        app_version {
            major
            minor
            patch
            type
            text
        }
        permissions {
            approved_scopes
            required_scopes
        }
    }
}
"""

# GraphQL query for fetching account information
ACCOUNT_QUERY = """
query {
    account {
        id
        name
        slug
        tier
        country_code
        first_day_of_the_week
        show_timeline_weekends
        sign_up_product_kind
        active_members_count
        logo
        plan {
            max_users
            period
            tier
            version
        }
    }
}
"""

# GraphQL query for fetching account roles
ACCOUNT_ROLES_QUERY = """
query {
    account_roles {
        id
        name
        roleType
    }
}
"""

# GraphQL query for fetching users
USERS_QUERY = """
query ($limit: Int!, $page: Int!) {
    users(limit: $limit, page: $page) {
        id
        name
        email
        enabled
        is_admin
        is_guest
        is_pending
        is_view_only
        created_at
        birthday
        country_code
        join_date
        location
        mobile_phone
        phone
        photo_original
        photo_thumb
        photo_tiny
        time_zone_identifier
        title
        url
        utc_hours_diff
        current_language
        account {
            id
        }
    }
}
"""

# GraphQL query for fetching boards
BOARDS_QUERY = """
query ($limit: Int!, $page: Int!) {
    boards(limit: $limit, page: $page) {
        id
        name
        description
        state
        board_kind
        board_folder_id
        workspace_id
        permissions
        item_terminology
        items_count
        updated_at
        url
        communication
        object_type_unique_key
        type
        creator {
            id
        }
        owners {
            id
        }
        subscribers {
            id
        }
        team_owners {
            id
        }
        team_subscribers {
            id
        }
        tags {
            id
            
        }
    }
}
"""

# GraphQL query for fetching custom activities
CUSTOM_ACTIVITIES_QUERY = """
query {
    custom_activity {
        id
        name
        type
        color
        icon_id
    }
}
"""

# GraphQL query for fetching board columns
BOARD_COLUMNS_QUERY = """
query ($board_ids: [ID!]) {
    boards(ids: $board_ids) {
        id
        columns {
            id
            title
            type
            archived
            description
            settings_str
            width
        }
    }
}
"""

# GraphQL query for fetching board views
BOARD_VIEWS_QUERY = """
query ($board_ids: [ID!]) {
    boards(ids: $board_ids) {
        id
        views {
            id
            name
            type
            settings_str
            view_specific_data_str
            source_view_id
            access_level
        }
    }
}
"""

# GraphQL query for fetching workspaces by IDs
WORKSPACES_QUERY = """
query ($ids: [ID!]) {
    workspaces(ids: $ids) {
        id
        name
        kind
        description
        created_at
        is_default_workspace
        state
        account_product {
            id
        }
        owners_subscribers {
            id
        }
        team_owners_subscribers {
            id
        }
        teams_subscribers {
            id
        }
        users_subscribers {
            id
        }
    }
}
"""

# GraphQL query for fetching webhooks by board ID
WEBHOOKS_QUERY = """
query ($board_id: ID!) {
    webhooks(board_id: $board_id) {
        id
        event
        board_id
        config
    }
}
"""

# GraphQL query for fetching updates
UPDATES_QUERY = """
query ($limit: Int!, $from_date: String, $to_date: String) {
    updates(limit: $limit, from_date: $from_date, to_date: $to_date) {
        id
        body
        text_body
        created_at
        updated_at
        edited_at
        creator_id
        item_id
        creator {
            id
        }
        item {
            id
        }
        assets {
            id
            name
            file_extension
            file_size
            public_url
            url
            url_thumbnail
            created_at
            original_geometry
            uploaded_by {
                id
            }
        }
        replies {
            id
            body
            text_body
            created_at
            updated_at
            creator_id
            creator {
                id
            }
        }
        likes {
            id
        }
        pinned_to_top {
            item_id
        }
        viewers {
            medium
            user_id
            user {
                id
            }
        }
    }
}
"""

# GraphQL query for fetching teams
TEAMS_QUERY = """
query {
    teams {
        id
        name
        picture_url
        users {
            id
            created_at
            phone
        }
    }
}
"""

# GraphQL query for fetching tags
TAGS_QUERY = """
query {
    tags {
        id
        name
        color
    }
}
"""

# Maximum number of results per page
MAX_PAGE_SIZE = 100



"""Helper functions for the Anthropic source using common HTTP client."""

import logging
from typing import Any, Callable, Dict, Iterator, List, Optional

import requests

from ingestr.src.http_client import create_client

logger = logging.getLogger(__name__)


class AnthropicClient:
    """HTTP client for Anthropic Admin API."""

    def __init__(self, api_key: str):
        self.api_key = api_key
        self.base_url = "https://api.anthropic.com/v1"
        self.headers = {
            "anthropic-version": "2023-06-01",
            "x-api-key": api_key,
            "User-Agent": "ingestr/1.0.0 (https://github.com/bruin-data/ingestr)",
        }
        # Create client with retry logic for common error codes
        self.client = create_client(retry_status_codes=[429, 502, 503, 504])

    def get(
        self, path: str, params: Optional[Dict[str, Any]] = None
    ) -> requests.Response:
        """Make a GET request to the Anthropic API."""
        url = f"{self.base_url}/{path}"
        return self.client.get(url, headers=self.headers, params=params)

    def fetch_paginated(
        self,
        path: str,
        params: Optional[Dict[str, Any]] = None,
        flatten_func: Optional[Callable] = None,
        handle_404: bool = False,
    ) -> Iterator[Dict[str, Any]]:
        """
        Fetch paginated data from the Anthropic API.

        Args:
            path: API endpoint path
            params: Query parameters
            flatten_func: Optional function to flatten records
            handle_404: If True, treat 404 as empty result instead of error

        Yields:
            Flattened records
        """
        if params is None:
            params = {}

        # Make a copy to avoid modifying the original
        params = dict(params)

        has_more = True
        next_page = None

        while has_more:
            current_params = dict(params)

            if next_page:
                current_params["page"] = next_page
                # Remove limit from subsequent requests as page cursor includes it
                current_params.pop("limit", None)

            try:
                response = self.get(path, current_params)
                response.raise_for_status()

                data = response.json()

                # Process each record
                for record in data.get("data", []):
                    if flatten_func:
                        yield flatten_func(record)
                    else:
                        yield record

                # Check for more pages
                has_more = data.get("has_more", False)
                next_page = data.get("next_page")

            except requests.exceptions.HTTPError as e:
                if e.response.status_code == 401:
                    raise ValueError(
                        "Invalid API key. Please ensure you're using an Admin API key "
                        "(starts with sk-ant-admin...) and have the necessary permissions."
                    )
                elif e.response.status_code == 404:
                    if handle_404:
                        logger.info(f"No data available for {path}")
                        break
                    else:
                        logger.info(f"No data found at {path}")
                        break
                elif e.response.status_code == 400:
                    error_msg = e.response.text
                    raise ValueError(f"Bad request: {error_msg}")
                else:
                    raise Exception(f"API request failed: {e}")
            except Exception as e:
                raise Exception(f"Failed to fetch data: {e}")

    def fetch_single(self, path: str) -> Optional[Dict[str, Any]]:
        """
        Fetch a single resource from the API.

        Args:
            path: API endpoint path

        Returns:
            The resource data or None if not found
        """
        try:
            response = self.get(path)
            response.raise_for_status()
            return response.json()
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 401:
                raise ValueError("Invalid API key")
            elif e.response.status_code == 404:
                return None
            raise Exception(f"Failed to fetch resource: {e}")


def flatten_usage_record(record: Dict[str, Any]) -> Dict[str, Any]:
    """
    Flatten a nested Claude Code usage record.

    Args:
        record: Nested record from API

    Returns:
        Flattened record
    """

    # Extract actor information
    actor = record.get("actor", {})
    actor_type = actor.get("type", "")
    actor_id = ""
    if actor_type == "user_actor":
        actor_id = actor.get("email_address", "")
    elif actor_type == "api_actor":
        actor_id = actor.get("api_key_name", "")

    # Start with base fields
    flattened = {
        "date": record.get("date"),
        "actor_type": actor_type,
        "actor_id": actor_id,
        "organization_id": record.get("organization_id"),
        "customer_type": record.get("customer_type"),
        "terminal_type": record.get("terminal_type"),
    }

    # Extract core metrics
    core_metrics = record.get("core_metrics", {})
    flattened.update(
        {
            "num_sessions": core_metrics.get("num_sessions", 0),
            "lines_added": core_metrics.get("lines_of_code", {}).get("added", 0),
            "lines_removed": core_metrics.get("lines_of_code", {}).get("removed", 0),
            "commits_by_claude_code": core_metrics.get("commits_by_claude_code", 0),
            "pull_requests_by_claude_code": core_metrics.get(
                "pull_requests_by_claude_code", 0
            ),
        }
    )

    # Extract tool actions (flatten nested structure)
    tool_actions = record.get("tool_actions", {})
    for tool, actions in tool_actions.items():
        flattened[f"{tool}_accepted"] = actions.get("accepted", 0)
        flattened[f"{tool}_rejected"] = actions.get("rejected", 0)

    # Extract model breakdown and aggregate totals
    model_breakdown = record.get("model_breakdown", [])
    total_input_tokens = 0
    total_output_tokens = 0
    total_cache_read_tokens = 0
    total_cache_creation_tokens = 0
    total_estimated_cost_cents = 0
    models_used = []

    for model_info in model_breakdown:
        model_name = model_info.get("model", "")
        if model_name:
            models_used.append(model_name)

        tokens = model_info.get("tokens", {})
        total_input_tokens += tokens.get("input", 0)
        total_output_tokens += tokens.get("output", 0)
        total_cache_read_tokens += tokens.get("cache_read", 0)
        total_cache_creation_tokens += tokens.get("cache_creation", 0)

        cost = model_info.get("estimated_cost", {})
        if cost.get("currency") == "USD":
            total_estimated_cost_cents += cost.get("amount", 0)

    flattened.update(
        {
            "total_input_tokens": total_input_tokens,
            "total_output_tokens": total_output_tokens,
            "total_cache_read_tokens": total_cache_read_tokens,
            "total_cache_creation_tokens": total_cache_creation_tokens,
            "total_estimated_cost_cents": total_estimated_cost_cents,
            "models_used": ",".join(models_used) if models_used else None,
        }
    )

    return flattened


def fetch_claude_code_usage(
    api_key: str,
    date: str,
    limit: int = 100,
) -> Iterator[Dict[str, Any]]:
    """
    Fetch Claude Code usage data for a specific date.

    Args:
        api_key: Anthropic Admin API key
        date: Date in YYYY-MM-DD format
        limit: Number of records per page (max 1000)

    Yields:
        Flattened usage records
    """
    client = AnthropicClient(api_key)
    params = {"starting_at": date, "ending_at": date, "limit": min(limit, 1000)}

    for record in client.fetch_paginated(
        "organizations/usage_report/claude_code",
        params=params,
        flatten_func=flatten_usage_record,
        handle_404=True,
    ):
        yield record


def flatten_usage_report_record(record: Dict[str, Any]) -> Dict[str, Any]:
    """
    Flatten a usage report record.

    Args:
        record: Nested usage record from API

    Returns:
        Flattened record
    """

    # Start with base fields - ensure bucket is never None
    flattened = {
        "bucket": record.get("bucket", ""),
        "api_key_id": record.get("api_key_id", ""),
        "workspace_id": record.get("workspace_id", ""),
        "model": record.get("model", ""),
        "service_tier": record.get("service_tier", ""),
        "context_window": record.get("context_window", ""),
    }

    # Extract token counts
    tokens = record.get("tokens", {})
    flattened.update(
        {
            "uncached_input_tokens": tokens.get("uncached_input_tokens", 0),
            "cached_input_tokens": tokens.get("cached_input_tokens", 0),
            "cache_creation_tokens": tokens.get("cache_creation_tokens", 0),
            "output_tokens": tokens.get("output_tokens", 0),
        }
    )

    # Extract server tool usage
    server_tool_usage = record.get("server_tool_usage", {})
    flattened.update(
        {
            "web_search_usage": server_tool_usage.get("web_search_usage", 0),
            "code_execution_usage": server_tool_usage.get("code_execution_usage", 0),
        }
    )

    return flattened


def fetch_usage_report(
    api_key: str,
    starting_at: str,
    ending_at: str,
    bucket_width: str = "1d",
    limit: int = 100,
    group_by: Optional[List[str]] = None,
    models: Optional[List[str]] = None,
    service_tiers: Optional[List[str]] = None,
    context_window: Optional[List[str]] = None,
    api_key_ids: Optional[List[str]] = None,
    workspace_ids: Optional[List[str]] = None,
) -> Iterator[Dict[str, Any]]:
    """
    Fetch usage report from the messages endpoint.

    Args:
        api_key: Anthropic Admin API key
        starting_at: Start date in ISO 8601 format
        ending_at: End date in ISO 8601 format
        bucket_width: Bucket width (1m, 1h, 1d)
        limit: Number of results per page
        group_by: Fields to group by
        models: Filter by models
        service_tiers: Filter by service tiers
        context_window: Filter by context window
        api_key_ids: Filter by API key IDs
        workspace_ids: Filter by workspace IDs

    Yields:
        Flattened usage records
    """
    client = AnthropicClient(api_key)

    # Adjust limit based on bucket_width
    max_limit = 31 if bucket_width == "1d" else (168 if bucket_width == "1h" else 1440)

    params = {
        "starting_at": starting_at,
        "ending_at": ending_at,
        "bucket_width": bucket_width,
        "limit": min(limit, max_limit),
    }

    # Add optional filters
    if group_by:
        for i, field in enumerate(group_by):
            params[f"group_by[{i}]"] = field
    if models:
        for i, model in enumerate(models):
            params[f"models[{i}]"] = model
    if service_tiers:
        for i, tier in enumerate(service_tiers):
            params[f"service_tiers[{i}]"] = tier
    if context_window:
        for i, window in enumerate(context_window):
            params[f"context_window[{i}]"] = window
    if api_key_ids:
        for i, key_id in enumerate(api_key_ids):
            params[f"api_key_ids[{i}]"] = key_id
    if workspace_ids:
        for i, workspace_id in enumerate(workspace_ids):
            params[f"workspace_ids[{i}]"] = workspace_id

    for record in client.fetch_paginated(
        "organizations/usage_report/messages",
        params=params,
        flatten_func=flatten_usage_report_record,
        handle_404=True,
    ):
        yield record


def fetch_cost_report(
    api_key: str,
    starting_at: str,
    ending_at: str,
    group_by: Optional[List[str]] = None,
    workspace_ids: Optional[List[str]] = None,
    limit: int = 31,
) -> Iterator[Dict[str, Any]]:
    """
    Fetch cost report data.

    Args:
        api_key: Anthropic Admin API key
        starting_at: Start date in ISO 8601 format
        ending_at: End date in ISO 8601 format
        group_by: Fields to group by
        workspace_ids: Filter by workspace IDs
        limit: Number of results per page (max 31)

    Yields:
        Cost records
    """
    client = AnthropicClient(api_key)

    params = {
        "starting_at": starting_at,
        "ending_at": ending_at,
        "limit": min(limit, 31),  # Max 31 for cost reports
    }

    # Add optional filters
    if group_by:
        for i, field in enumerate(group_by):
            params[f"group_by[{i}]"] = field
    if workspace_ids:
        for i, workspace_id in enumerate(workspace_ids):
            params[f"workspace_ids[{i}]"] = workspace_id

    def flatten_cost_record(record: Dict[str, Any]) -> Dict[str, Any]:
        """Flatten cost record with defaults for nullable fields."""
        return {
            "bucket": record.get("bucket", ""),
            "workspace_id": record.get("workspace_id", ""),
            "description": record.get("description", ""),
            "amount_cents": record.get("amount_cents", 0),
        }

    for record in client.fetch_paginated(
        "organizations/cost_report",
        params=params,
        flatten_func=flatten_cost_record,
        handle_404=True,
    ):
        yield record


def fetch_organization_info(api_key: str) -> Optional[Dict[str, Any]]:
    """
    Fetch organization information.

    Args:
        api_key: Anthropic Admin API key

    Returns:
        Organization information
    """
    client = AnthropicClient(api_key)
    return client.fetch_single("organizations/me")


def fetch_workspaces(api_key: str, limit: int = 100) -> Iterator[Dict[str, Any]]:
    """
    Fetch all workspaces in the organization.

    Args:
        api_key: Anthropic Admin API key
        limit: Number of records per page

    Yields:
        Workspace records
    """
    client = AnthropicClient(api_key)
    params = {"limit": min(limit, 100)}

    for record in client.fetch_paginated("workspaces", params=params):
        yield record


def fetch_api_keys(api_key: str, limit: int = 100) -> Iterator[Dict[str, Any]]:
    """
    Fetch all API keys in the organization.

    Args:
        api_key: Anthropic Admin API key
        limit: Number of records per page

    Yields:
        API key records
    """
    client = AnthropicClient(api_key)
    params = {"limit": min(limit, 100)}

    for record in client.fetch_paginated("api_keys", params=params):
        yield record


def fetch_invites(api_key: str, limit: int = 100) -> Iterator[Dict[str, Any]]:
    """
    Fetch all pending invites.

    Args:
        api_key: Anthropic Admin API key
        limit: Number of records per page

    Yields:
        Invite records
    """
    client = AnthropicClient(api_key)
    params = {"limit": min(limit, 100)}

    for record in client.fetch_paginated("invites", params=params):
        yield record


def fetch_users(api_key: str, limit: int = 100) -> Iterator[Dict[str, Any]]:
    """
    Fetch all users in the organization.

    Args:
        api_key: Anthropic Admin API key
        limit: Number of records per page

    Yields:
        User records
    """
    client = AnthropicClient(api_key)
    params = {"limit": min(limit, 100)}

    for record in client.fetch_paginated("users", params=params):
        yield record


def fetch_workspace_members(
    api_key: str, workspace_id: Optional[str] = None, limit: int = 100
) -> Iterator[Dict[str, Any]]:
    """
    Fetch workspace members.

    Args:
        api_key: Anthropic Admin API key
        workspace_id: Optional workspace ID to filter by
        limit: Number of records per page

    Yields:
        Workspace member records
    """
    client = AnthropicClient(api_key)
    params: Dict[str, Any] = {"limit": min(limit, 100)}
    if workspace_id:
        params["workspace_id"] = workspace_id

    for record in client.fetch_paginated("workspace_members", params=params):
        yield record
